Package: resultcheck
Title: Result Stability Checks for Empirical R Projects
Version: 0.2.1
Authors@R: 
    person("Dianyi", "Yang", , "dianyi.yang@politics.ox.ac.uk", role = c("aut", "cre", "ctb"), comment = c(ORCID = "0009-0004-4652-3429"))
Description: 
    Lightweight helpers for checking whether empirical results remain 
    substantively unchanged across code revisions, platform differences, 
    and package updates. The package supports regression-style testing of 
    derived datasets, statistical model outputs, tables, and plots, helping 
    researchers detect unintended result drift early and distinguish material 
    from non-material changes in empirical workflows.
License: MIT + file LICENSE
Encoding: UTF-8
Imports: withr, rprojroot, yaml
Suggests: broom, knitr, rmarkdown, rstudioapi, testthat (>= 3.0.0),
        waldo
URL: https://github.com/kv9898/resultcheck/,
        https://kv9898.github.io/resultcheck/
BugReports: https://github.com/kv9898/resultcheck/issues
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-08 12:38:06 UTC; Dianyi
Author: Dianyi Yang [aut, cre, ctb] (ORCID:
    <https://orcid.org/0009-0004-4652-3429>)
Maintainer: Dianyi Yang <dianyi.yang@politics.ox.ac.uk>
Repository: CRAN
Date/Publication: 2026-05-08 16:40:02 UTC
Built: R 4.6.0; ; 2026-05-08 20:32:07 UTC; unix

# Generated by roxygen2: do not edit by hand

export(cleanup_sandbox)
export(find_root)
export(run_in_sandbox)
export(setup_sandbox)
export(snapshot)
export(with_example)
importFrom(grDevices,dev.off)
importFrom(grDevices,pdf)
importFrom(stats,lm)
importFrom(utils,getSrcref)
importFrom(utils,str)
importFrom(yaml,read_yaml)

# resultcheck

<!-- badges: start -->

[![R-CMD_CI_Tests_Badge](https://github.com/kv9898/resultcheck/actions/workflows/ci-tests.yml/badge.svg)](https://github.com/kv9898/resultcheck/actions/workflows/ci-tests.yml)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/resultcheck)](https://cran.r-project.org/package=resultcheck)
[![CRAN_Downloads_Badge](https://cranlogs.r-pkg.org/badges/resultcheck)](https://cran.r-project.org/package=resultcheck)

<!-- badges: end -->

Result Stability Checks for Empirical R Projects

## Overview

`resultcheck` provides lightweight helpers for checking whether empirical results remain unchanged across code revisions, platform differences, and package updates. Call `snapshot()` on key outputs (models, tables, derived datasets) in your analysis scripts to detect unintended result drift automatically during CI or local testing.

## Installation

### Latest Stable Version

```r
install.packages("resultcheck")
```

### Latest Development Version (Unstable)

```r
# install.packages("devtools")
devtools::install_github("kv9898/resultcheck")
```

## Workflow

The package supports a two-phase workflow:

1. **Interactive development** — run your analysis script and call `snapshot()` on objects you care about. On first run the snapshot is saved as a human-readable `.md` file. On subsequent interactive runs, differences are shown and you are prompted to update.

2. **Automated testing** — wrap your script in `setup_sandbox()` / `run_in_sandbox()` / `cleanup_sandbox()`. Inside `run_in_sandbox()`, `snapshot()` switches to *testing mode*: it errors immediately if a snapshot is missing or has changed, making the test fail.

## Integrated Example

`with_example()` can generate this layout for documentation/testing under `tempdir()`:

```
myproject/
├── _resultcheck.yml
├── analysis.R
└── tests/
    ├── _resultcheck_snaps/
    │   └── analysis/
    │       ├── model.md
    │       └── model_mismatch.md
    └── testthat/
        └── test-analysis.R
```

### `analysis.R` — snapshot key results

```r
model <- lm(mpg ~ wt, data = mtcars)
resultcheck::snapshot(model, "model")
```

### `tests/testthat/test-analysis.R` — automated test

```r
library(testthat)
library(resultcheck)

test_that("analysis produces stable results", {
  sandbox <- setup_sandbox()
  on.exit(cleanup_sandbox(sandbox), add = TRUE)

  expect_true(run_in_sandbox("analysis.R", sandbox))
})
```

To try this quickly without creating files in your current project:

```r
resultcheck::with_example({
  sandbox <- setup_sandbox()
  on.exit(cleanup_sandbox(sandbox), add = TRUE)
  stopifnot(isTRUE(run_in_sandbox("analysis.R", sandbox)))
})
```

---

## Function Reference

### `snapshot(value, name, script_name = NULL, method = NULL)`

Creates or verifies a snapshot of any R object.

- **First interactive run**: saves the object as a human-readable `.md` file under `tests/_resultcheck_snaps/<script>/` at the project root by default.
- **Subsequent interactive runs**: shows a diff and prompts to update.
- **Inside `run_in_sandbox()`**: errors if the snapshot is missing or doesn't match.
- **When writing snapshots interactively**: warns and shows the exact output path.

You can override the default snapshot directory in `_resultcheck.yml`:

```yaml
snapshot:
  dir: "custom/snapshots/path"
  method: "print + str"
  method_defaults_file: "snapshot-method-overrides.R"
  method_by_class:
    lm: "summary"
```

The `method` argument controls how the object is serialized:

| Value | Behavior |
|-------|-----------|
| `NULL` (default) | Captures both `print()` and `str()` |
| `print` | Only `print()` output is captured |
| `str` | Only `str()` output is captured |
| `length` | Any callable function can be used |
| `stats::coef` | Namespaced callables are supported |
| `list(print = print, summary = summary)` | Runs multiple functions in order |

When a list of methods is used, section headers are taken directly from the
method names (or list names), e.g. `## print`, `## summary`.

If `method` is omitted, defaults are resolved as:

1. class override from `snapshot.method_by_class`,
2. global default from `snapshot.method`,
3. fallback to `list(print = print, str = str)`.

Config methods are parsed from strings in `_resultcheck.yml`, so values like
`"print + str"` or `"stats::coef"` are converted to callable methods.

You can also define class defaults in a separate R file:

```r
# snapshot-method-overrides.R
method_by_class <- list(
  lm = "summary",
  glm = "summary"
)
```

Snapshots are plain text and intended to be committed to version control.

### `setup_sandbox(files = NULL, temp_base = NULL)`

Creates a temporary directory and copies the listed files and/or directories into it, preserving their path structure relative to the project root. Directories are copied recursively. Snapshot files do not need to be listed.

### `run_in_sandbox(script_path, sandbox = NULL, ...)`

Runs an R script inside the sandbox. The working directory is set to the sandbox, but `find_root()` and `snapshot()` automatically resolve back to the original project root so snapshots are found correctly.

Returns `TRUE` invisibly on success, so you can use `expect_true(run_in_sandbox(...))` directly in testthat.

### `cleanup_sandbox(sandbox = NULL, force = TRUE)`

Removes the sandbox directory. Omit the argument to clean up the most recently created sandbox.

### `find_root(start_path = NULL)`

Locates the project root by searching upward for a `_resultcheck.yml` (or legacy `resultcheck.yml`), `.Rproj`, or `.git` marker. Called automatically by `snapshot()` and `setup_sandbox()`.

Place an empty `_resultcheck.yml` at your project root to make detection reliable:

```yaml
# _resultcheck.yml
```

### `with_example(code, mismatch = FALSE)`

Creates a temporary example project in `tempdir()`, sets the working directory there while evaluating `code`, then cleans up automatically.

---

## License

MIT © Dianyi Yang

# Built-in class-based snapshot method defaults.
# Define `method_by_class` as a named list where each value is a method
# expression string (for example "summary") or a method spec.
#
# These defaults provide sensible snapshotting methods for common R classes:
# - "print"  : uses the object's print method
# - "str"    : uses str() to show structure
# - "summary": uses summary() (best for statistical models like lm, glm)
# - etc. (any function accessible via name or pkg::name syntax)
method_by_class <- list(
  # Linear models - broom::tidy gives useful coefficients table
  lm = c("broom::tidy", "broom::glance", "broom::augment"),
  glm = c("broom::tidy", "broom::glance", "broom::augment"),
  aov = c("broom::tidy", "broom::glance"),
  manova = "broom::tidy",

  # Extended linear models (broom supports tidy + glance for all)
  lm.beta = "broom::tidy",
  lmodel2 = c("broom::tidy", "broom::glance"),
  biglm = c("broom::tidy", "broom::glance"),
  speedlm = c("broom::tidy", "broom::glance", "broom::augment"),

  # Robust linear models
  rlm = c("broom::tidy", "broom::glance", "broom::augment"),
  lmrob = c("broom::tidy", "broom::glance", "broom::augment"),
  lmRob = c("broom::tidy", "broom::glance", "broom::augment"),
  glmrob = "broom::tidy",
  glmRob = c("broom::tidy", "broom::glance", "broom::augment"),

  # Regularized/glmnet models
  glmnet = c("broom::tidy", "broom::glance"),
  cv.glmnet = c("broom::tidy", "broom::glance"),

  # Nonlinear models
  nls = c("broom::tidy", "broom::glance", "broom::augment"),
  nlrq = c("broom::tidy", "broom::glance", "broom::augment"),

  # Mixed effects / multilevel
  lme = c("broom::tidy", "broom::glance"),
  merMod = c("broom::tidy", "broom::glance"),
  lmerMod = c("broom::tidy", "broom::glance"),
  glmerMod = c("broom::tidy", "broom::glance"),
  nlmerMod = c("broom::tidy", "broom::glance"),

  # GAM models
  gam = c("broom::tidy", "broom::glance", "broom::augment"),
  Gam = c("broom::tidy", "broom::glance"),

  # Survival models
  coxph = c("broom::tidy", "broom::glance", "broom::augment"),
  survreg = c("broom::tidy", "broom::glance", "broom::augment"),
  aareg = c("broom::tidy", "broom::glance"),
  cch = c("broom::tidy", "broom::glance"),
  crr = c("broom::tidy", "broom::glance"),
  survfit = c("broom::tidy", "broom::glance"),
  survdiff = c("broom::tidy", "broom::glance"),
  survexp = c("broom::tidy", "broom::glance"),
  muhaz = c("broom::tidy", "broom::glance"),
  pyears = c("broom::tidy", "broom::glance"),

  # Time series / ARIMA
  Arima = c("broom::tidy", "broom::glance"),
  acf = "broom::tidy",
  spec = "broom::tidy",
  ts = "broom::tidy",
  zoo = "broom::tidy",

  # Bayesian models
  stanfit = c("broom::tidy", "broom::glance"),
  bamlss = c("broom::tidy", "broom::glance"),

  # Panel data models
  plm = c("broom::tidy", "broom::glance", "broom::augment"),

  # Count / GLM extensions
  negbin = c("broom::tidy", "broom::glance"),
  speedglm = c("broom::tidy", "broom::glance", "broom::augment"),

  # GEE models
  geeglm = c("broom::tidy", "broom::glance"),

  # Ordinal models
  polr = c("broom::tidy", "broom::glance", "broom::augment"),
  clm = c("broom::tidy", "broom::glance", "broom::augment"),
  clmm = c("broom::tidy", "broom::glance"),
  svyolr = c("broom::tidy", "broom::glance"),

  # Multinomial models
  multinom = c("broom::tidy", "broom::glance"),
  mlogit = c("broom::tidy", "broom::glance", "broom::augment"),

  # Beta regression
  betareg = c("broom::tidy", "broom::glance", "broom::augment"),

  # Cure regression
  crr = c("broom::tidy", "broom::glance"),

  # Cluster analysis
  kmeans = c("broom::tidy", "broom::glance", "broom::augment"),
  Mclust = c("broom::tidy", "broom::glance", "broom::augment"),
  pam = c("broom::tidy", "broom::glance", "broom::augment"),

  # PCA / dimension reduction
  prcomp = c("broom::tidy", "broom::augment"),
  factanal = c("broom::tidy", "broom::glance", "broom::augment"),

  # DEA / stochastic frontier
  felm = c("broom::tidy", "broom::glance", "broom::augment"),
  ridgelm = c("broom::tidy", "broom::glance"),

  # Quantile regression
  rq = c("broom::tidy", "broom::glance", "broom::augment"),
  rqs = c("broom::tidy", "broom::glance", "broom::augment"),

  # GMM
  gmm = c("broom::tidy", "broom::glance"),

  # Instrumental variables
  ivreg = c("broom::tidy", "broom::glance", "broom::augment"),

  # Meta-analysis
  rma = c("broom::tidy", "broom::glance", "broom::augment"),

  # Network models
  ergm = c("broom::tidy", "broom::glance"),
  btergm = "broom::tidy",
  sarlm = c("broom::tidy", "broom::glance", "broom::augment"),
  Sarlm = c("broom::tidy", "broom::glance", "broom::augment"),

  # Marginal effects
  mfx = c("broom::tidy", "broom::glance", "broom::augment"),
  betamfx = c("broom::tidy", "broom::glance", "broom::augment"),
  logitmfx = c("broom::tidy", "broom::glance", "broom::augment"),
  probitmfx = c("broom::tidy", "broom::glance", "broom::augment"),
  negbinmfx = c("broom::tidy", "broom::glance", "broom::augment"),
  poissonmfx = c("broom::tidy", "broom::glance", "broom::augment"),
  margins = c("broom::tidy", "broom::glance", "broom::augment"),

  # Fixest models
  fixest = c("broom::tidy", "broom::glance", "broom::augment"),

  # Survey models
  svyglm = c("broom::tidy", "broom::glance"),

  # Latent class
  poLCA = c("broom::tidy", "broom::glance", "broom::augment"),

  # Joint models
  mjoint = c("broom::tidy", "broom::glance", "broom::augment"),

  # DRM / DRC
  drc = c("broom::tidy", "broom::glance", "broom::augment"),

  # GARCH
  garch = c("broom::tidy", "broom::glance"),

  # ANOVA extensions
  anova = c("broom::tidy", "broom::glance"),
  aovlist = "broom::tidy",

  # Coefficient tests
  coeftest = c("broom::tidy", "broom::glance"),

  # Tests and diagnostics
  htest = c("broom::tidy", "broom::glance", "broom::augment"),
  durbinWatsonTest = c("broom::tidy", "broom::glance"),

  # Confidence intervals
  confint.glht = "broom::tidy",
  glht = "broom::tidy",
  summary.glht = "broom::tidy",

  # Bootstrapping
  boot = "broom::tidy",

  # Statistical tests
  pairwise.htest = "broom::tidy",
  leveneTest = "broom::tidy",
  TukeyHSD = "broom::tidy",

  # Table/matrix summaries
  table = "broom::tidy",
  ftable = "broom::tidy",
  confusionMatrix = "broom::tidy",

  # Variance models
  varest = c("broom::tidy", "broom::glance"),
  lavaan = c("broom::tidy", "broom::glance"),
  ref.grid = "broom::tidy",
  emmGrid = "broom::tidy",
  summary_emm = "broom::tidy",
  lsmobj = "broom::tidy",

  # Binomial design
  binDesign = c("broom::tidy", "broom::glance"),
  binWidth = "broom::tidy",
  power.htest = "broom::tidy"
)

# Read input data
data <- read.csv("data/income.csv")

# Fit a linear model
model <- lm(income ~ age + education, data = data)

# Snapshot the model to detect unexpected result changes across code revisions.
# In interactive use: warns and prompts to update when differences are found.
# Inside run_in_sandbox(): errors if the snapshot is missing or doesn't match.
resultcheck::snapshot(model, "income_model")

# Write model summary to an output file
dir.create("output", showWarnings = FALSE)
write.csv(
  as.data.frame(coef(summary(model))),
  "output/model_summary.csv"
)

#' Find Project Root Directory
#'
#' Finds the root directory of the current R project using various heuristics.
#' The function searches for markers like \code{_resultcheck.yml} (preferred),
#' \code{resultcheck.yml} (legacy), \code{.Rproj} files,
#' or a \code{.git} directory. When running inside a sandbox created by
#' \code{setup_sandbox()}, it will search from the original working directory.
#'
#' @param start_path Optional. The directory to start searching from.
#'   If NULL (default), uses the current working directory or the stored
#'   original working directory if in a sandbox.
#'
#' @return The path to the project root directory.
#'
#' @export
#'
#' @examples
#' with_example({
#'   root <- find_root()
#'   print(root)
#' })
find_root <- function(start_path = NULL) {
  if (!requireNamespace("rprojroot", quietly = TRUE)) {
    stop("Package 'rprojroot' is required but not installed. ",
         "Please install it with: install.packages('rprojroot')")
  }
  
  # Determine start_path if not provided
  if (is.null(start_path)) {
    start_path <- get_start_path_for_find_root()
  }
  
  # Validate and normalize start_path
  if (!dir.exists(start_path)) {
    stop("Start path does not exist: ", start_path, call. = FALSE)
  }
  
  # Normalize the path to avoid issues with rprojroot
  start_path <- normalizePath(start_path, winslash = "/", mustWork = TRUE)
  
  # Define criteria for finding project root
  # Try multiple criteria in order of preference
  criteria <- rprojroot::has_file("_resultcheck.yml") |
    rprojroot::has_file("resultcheck.yml") |
    rprojroot::has_file_pattern("[.]Rproj$") |
    rprojroot::is_git_root
  
  tryCatch({
    root <- rprojroot::find_root(criteria, path = start_path)
    return(root)
  }, error = function(e) {
    stop("Could not find project root from path: ", start_path, ". ",
         "Please ensure you are in a project directory ",
         "with either a _resultcheck.yml (or legacy resultcheck.yml), ",
         ".Rproj file, or .git directory. ",
         "Original error: ", e$message, call. = FALSE)
  })
}


#' Get start path for find_root
#' 
#' Helper function to determine the starting path for find_root().
#' Checks for stored sandbox WD first, then falls back to getwd().
#' 
#' @return Character path to start searching from
#' @keywords internal
get_start_path_for_find_root <- function() {
  # Check if we have a stored original working directory from sandbox
  if (exists(".resultcheck_original_wd", envir = .resultcheck_env)) {
    start_path <- .resultcheck_env$.resultcheck_original_wd
    # Validate the stored path still exists
    if (!is.null(start_path) && dir.exists(start_path)) {
      return(start_path)
    } else if (!is.null(start_path)) {
      warning("Stored original WD no longer exists: ", start_path, 
              ". Falling back to getwd().", immediate. = TRUE)
    }
  }
  
  # Try to get current working directory
  start_path <- tryCatch(getwd(), error = function(e) NULL)
  
  # If getwd() fails or returns NULL
  if (is.null(start_path) || length(start_path) == 0 || start_path == "" || is.na(start_path)) {
    stop("Could not determine current working directory. ",
         "This may happen if the current directory has been deleted. ",
         "Please ensure you are in a valid directory.",
         call. = FALSE)
  }
  
  return(start_path)
}


#' Detect the calling script name
#'
#' Tries to identify the script file that is calling snapshot().
#' Uses \code{rstudioapi::getSourceEditorContext()$path} when running in RStudio,
#' falling back to walking the call stack for source references, and finally
#' to \code{"interactive"}.
#'
#' @return Character string with the detected script basename (without path).
#' @keywords internal
detect_script_name <- function() {
  # RStudio: use the active source editor path
  path <- tryCatch(
    rstudioapi::getSourceEditorContext()$path,
    error = function(e) NULL
  )
  if (is.character(path) && nzchar(path) && path != "") {
    return(basename(path))
  }

  # Fallback: walk the call stack for source references,
  # skipping any that point to files within the resultcheck package itself.
  # When loaded via pkgload (development), internal functions have source refs;
  # we find the package R directory from this function's own source filename.
  pkg_r_dir <- tryCatch({
    f <- utils::getSrcFilename(detect_script_name, full.names = TRUE)
    if (!is.null(f) && length(f) > 0L && nzchar(f[[1L]]))
      dirname(normalizePath(f[[1L]], mustWork = FALSE, winslash = "/"))
    else
      ""
  }, error = function(e) "")
  # Use a larger depth (20) than the old limit (10) to reliably skip several
  # internal resultcheck frames (detect_script_name, get_snapshot_path, snapshot,
  # run_in_sandbox, withr layers, ...) before reaching user-script frames.
  max_frames <- 20L
  for (i in seq_len(min(max_frames, sys.nframe()))) {
    srcref <- tryCatch(getSrcref(sys.call(-i)), error = function(e) NULL)
    if (is.null(srcref)) next
    srcfile <- attr(srcref, "srcfile")
    if (is.null(srcfile) || is.null(srcfile$filename) || !nzchar(srcfile$filename)) next
    if (nzchar(pkg_r_dir)) {
      fn <- tryCatch(
        normalizePath(srcfile$filename, mustWork = FALSE, winslash = "/"),
        error = function(e) srcfile$filename
      )
      if (startsWith(fn, pkg_r_dir)) next
    }
    return(basename(srcfile$filename))
  }

  "interactive"
}

#' Get Snapshot File Path
#'
#' Constructs the path to a snapshot file within the project's snapshot directory.
#' Snapshot files are stored under \code{tests/_resultcheck_snaps/} by default,
#' organized by script name. This location can be overridden with
#' \code{snapshot.dir} in \code{_resultcheck.yml}.
#'
#' @param name Character. The name of the snapshot (without extension).
#' @param script_name Optional. The name of the script file creating the snapshot.
#'   If NULL, detects from the active RStudio source editor via
#'   \code{rstudioapi::getSourceEditorContext()$path}, falling back to the call
#'   stack, then to \code{"interactive"}.
#' @param ext Character. The file extension for the snapshot file (default: "md").
#'
#' @return The full path to the snapshot file.
#'
#' @keywords internal
get_snapshot_path <- function(name, script_name = NULL, ext = "md") {
  root <- find_root()
  config <- read_resultcheck_config()

  # If script_name not provided, try to detect the source file
  if (is.null(script_name)) {
    script_name <- detect_script_name()
  }

  # Clean up script name (remove extension)
  script_name <- sub("\\.[Rr]$", "", script_name)
  
  snapshot_base <- config[["snapshot"]][["dir"]]
  if (!is.character(snapshot_base) ||
      length(snapshot_base) != 1L ||
      is.na(snapshot_base) ||
      trimws(snapshot_base) == "") {
    snapshot_base <- "tests/_resultcheck_snaps"
  }
  snapshot_base <- trimws(snapshot_base)
  is_absolute <- startsWith(snapshot_base, "/") ||
    grepl("^[A-Za-z]:[/\\\\]", snapshot_base)
  snapshot_root <- if (is_absolute) snapshot_base else file.path(root, snapshot_base)

  # Construct snapshot directory path
  snapshot_dir <- file.path(snapshot_root, script_name)
  
  # Create directory if it doesn't exist
  if (!dir.exists(snapshot_dir)) {
    dir.create(snapshot_dir, recursive = TRUE, showWarnings = FALSE)
  }
  
  # Return full path to snapshot file
  snapshot_file <- file.path(snapshot_dir, paste0(name, ".", ext))
  return(snapshot_file)
}


# Marker string that users can place on any line of a snapshot file to indicate
# that the line's value is known to vary across platforms and should be ignored
# during comparison.  The entire line must consist of exactly this string.
IGNORED_MARKER <- "[ignored]"


#' Read resultcheck Configuration
#'
#' Reads configuration settings from the \code{_resultcheck.yml} file located
#' at the project root (falling back to legacy \code{resultcheck.yml} if
#' needed). Returns an empty list if neither file exists or parsing fails.
#'
#' @return A named list of configuration values, or an empty list.
#'
#' @keywords internal
read_resultcheck_config <- function() {
  tryCatch({
    root <- find_root()
    config_path <- file.path(root, "_resultcheck.yml")
    if (!file.exists(config_path)) {
      legacy_path <- file.path(root, "resultcheck.yml")
      config_path <- if (file.exists(legacy_path)) legacy_path else config_path
    }
    if (!file.exists(config_path)) return(list())
    config <- yaml::read_yaml(config_path)
    if (is.null(config)) {
      config <- list()
    }
    config[["snapshot"]] <- normalize_snapshot_config(config[["snapshot"]], root = root)
    config
  }, error = function(e) list())
}


DEFAULT_SNAPSHOT_METHODS <- list(
  print = base::print,
  str = utils::str
)

SNAPSHOT_METHOD_DEFAULTS_FILE <- "snapshot-method-defaults.R"

resolve_snapshot_defaults_file <- function(path, root) {
  if (!is.character(path) || length(path) != 1L || is.na(path)) {
    stop("snapshot.method_defaults_file must be a non-empty string.", call. = FALSE)
  }
  path <- trimws(path)
  if (path == "") {
    stop("snapshot.method_defaults_file must be a non-empty string.", call. = FALSE)
  }

  is_absolute <- startsWith(path, "/") || grepl("^[A-Za-z]:[\\\\/]", path)
  if (is_absolute) {
    return(path)
  }
  file.path(root, path)
}

resolve_config_method_token <- function(token, arg_name) {
  token <- trimws(token)
  if (!nzchar(token)) {
    stop("`", arg_name, "` contains an empty method token.", call. = FALSE)
  }

  if (grepl(":::", token, fixed = TRUE)) {
    parts <- strsplit(token, ":::", fixed = TRUE)[[1]]
    if (length(parts) != 2L || any(trimws(parts) == "")) {
      stop("Invalid namespaced method token `", token, "` in `", arg_name, "`.", call. = FALSE)
    }
    pkg <- trimws(parts[1L])
    fn_name <- trimws(parts[2L])
    fn <- tryCatch(
      utils::getFromNamespace(fn_name, pkg),
      error = function(e) {
        stop(
          "Could not resolve `", token, "` in `", arg_name, "`: ",
          conditionMessage(e),
          call. = FALSE
        )
      }
    )
    if (!is.function(fn)) {
      stop("`", token, "` in `", arg_name, "` does not resolve to a function.", call. = FALSE)
    }
    return(list(fn = fn, label = token))
  }

  if (grepl("::", token, fixed = TRUE)) {
    parts <- strsplit(token, "::", fixed = TRUE)[[1]]
    if (length(parts) != 2L || any(trimws(parts) == "")) {
      stop("Invalid namespaced method token `", token, "` in `", arg_name, "`.", call. = FALSE)
    }
    pkg <- trimws(parts[1L])
    fn_name <- trimws(parts[2L])
    fn <- tryCatch(
      getExportedValue(pkg, fn_name),
      error = function(e) {
        stop(
          "Could not resolve `", token, "` in `", arg_name, "`: ",
          conditionMessage(e),
          call. = FALSE
        )
      }
    )
    if (!is.function(fn)) {
      stop("`", token, "` in `", arg_name, "` does not resolve to a function.", call. = FALSE)
    }
    return(list(fn = fn, label = token))
  }

  fn <- tryCatch(
    get(token, mode = "function", inherits = TRUE),
    error = function(e) {
      stop(
        "Could not resolve method `", token, "` in `", arg_name, "`: ",
        conditionMessage(e),
        call. = FALSE
      )
    }
  )
  if (!is.function(fn)) {
    stop("`", token, "` in `", arg_name, "` does not resolve to a function.", call. = FALSE)
  }
  list(fn = fn, label = token)
}

parse_config_method_expression <- function(method, arg_name = "snapshot.method") {
  if (!is.character(method)) {
    stop("`", arg_name, "` must be a character method expression or vector.", call. = FALSE)
  }
  if (length(method) == 0L) {
    stop("`", arg_name, "` must include at least one method token.", call. = FALSE)
  }

  parts <- unlist(strsplit(method, "\\+", perl = TRUE), use.names = FALSE)
  parts <- trimws(parts)
  parts <- parts[!is.na(parts) & nzchar(parts)]
  if (length(parts) == 0L) {
    stop("`", arg_name, "` must include at least one method token.", call. = FALSE)
  }

  method_specs <- lapply(parts, function(token) resolve_config_method_token(token, arg_name))
  labels <- vapply(method_specs, `[[`, character(1), "label")
  dedup_idx <- !duplicated(labels)
  method_specs[dedup_idx]
}

coerce_snapshot_methods_from_config <- function(method, arg_name = "snapshot.method") {
  if (is.character(method)) {
    return(parse_config_method_expression(method, arg_name = arg_name))
  }
  coerce_snapshot_methods(method, arg_name = arg_name)
}

coerce_class_override_map <- function(x, source = "snapshot.method_by_class",
                                      skip_errors = FALSE) {
  if (is.null(x)) return(list())
  if (!is.list(x)) {
    stop(source, " must be a named list mapping class names to method values.", call. = FALSE)
  }
  if (length(x) == 0L) return(list())

  nms <- names(x)
  if (is.null(nms) || anyNA(nms) || any(trimws(nms) == "")) {
    stop(source, " must be a named list mapping class names to method values.", call. = FALSE)
  }

  out <- list()
  for (nm in nms) {
    class_name <- trimws(nm)
    if (!nzchar(class_name)) next
    if (skip_errors) {
      entry <- tryCatch(
        coerce_snapshot_methods_from_config(
          x[[nm]],
          arg_name = paste0(source, "$", class_name)
        ),
        error = function(e) NULL
      )
      if (!is.null(entry)) out[[class_name]] <- entry
    } else {
      out[[class_name]] <- coerce_snapshot_methods_from_config(
        x[[nm]],
        arg_name = paste0(source, "$", class_name)
      )
    }
  }
  out
}

eval_character_literal <- function(expr, source_name, label) {
  if (is.character(expr)) {
    return(expr)
  }
  if (is.call(expr) && identical(expr[[1L]], quote(c))) {
    args <- as.list(expr)[-1L]
    vals <- vapply(args, function(arg) {
      if (!is.character(arg) || length(arg) != 1L) {
        stop(
          source_name, " has unsupported value for ", label,
          ". Use character method expressions only in defaults files.",
          call. = FALSE
        )
      }
      arg
    }, character(1L))
    return(unname(vals))
  }
  stop(
    source_name, " has unsupported value for ", label,
    ". Use character method expressions only in defaults files.",
    call. = FALSE
  )
}

eval_class_map_expr <- function(expr, source_name) {
  if (!is.call(expr) || !identical(expr[[1L]], quote(list))) {
    stop(
      source_name,
      " must define `method_by_class` as a named list of character method expressions.",
      call. = FALSE
    )
  }
  args <- as.list(expr)[-1L]
  nms <- names(args)
  if (length(args) == 0L) return(list())
  if (is.null(nms) || anyNA(nms) || any(trimws(nms) == "")) {
    stop(
      source_name,
      " must define a named list mapping class names to method expressions.",
      call. = FALSE
    )
  }

  out <- list()
  for (i in seq_along(args)) {
    class_name <- trimws(nms[[i]])
    out[[class_name]] <- eval_character_literal(
      args[[i]],
      source_name = source_name,
      label = paste0("`", class_name, "`")
    )
  }
  out
}

extract_class_map_expr <- function(exprs) {
  if (length(exprs) == 0L) return(NULL)

  for (expr in exprs) {
    if (is.call(expr) &&
        (identical(expr[[1L]], quote(`<-`)) || identical(expr[[1L]], quote(`=`))) &&
        is.symbol(expr[[2L]]) &&
        as.character(expr[[2L]]) %in% c("snapshot_method_by_class", "method_by_class")) {
      return(expr[[3L]])
    }
  }

  if (length(exprs) == 1L) return(exprs[[1L]])
  NULL
}

read_method_defaults_from_r_file <- function(path, source_name, skip_errors = FALSE) {
  if (!is.character(path) || length(path) != 1L || is.na(path) || trimws(path) == "") {
    return(list())
  }
  if (!file.exists(path)) {
    stop("Configured snapshot defaults file does not exist: ", path, call. = FALSE)
  }

  exprs <- parse(file = path, keep.source = FALSE)
  class_expr <- extract_class_map_expr(exprs)
  if (is.null(class_expr)) {
    return(list())
  }
  class_map <- eval_class_map_expr(class_expr, source_name = source_name)

  coerce_class_override_map(class_map, source = source_name, skip_errors = skip_errors)
}

normalize_snapshot_config <- function(snapshot_cfg, root) {
  defaults <- list(
    method = NULL,
    method_by_class = list()
  )
  if (is.null(snapshot_cfg)) {
    snapshot_cfg <- list()
  }
  if (!is.list(snapshot_cfg)) {
    stop("`snapshot` configuration must be a list.", call. = FALSE)
  }

  cfg <- utils::modifyList(defaults, snapshot_cfg)

  pkg_defaults_path <- system.file("extdata", SNAPSHOT_METHOD_DEFAULTS_FILE, package = "resultcheck")
  pkg_class_map <- if (nzchar(pkg_defaults_path)) {
    read_method_defaults_from_r_file(
      pkg_defaults_path,
      source_name = "built-in snapshot method defaults",
      skip_errors = TRUE
    )
  } else {
    list()
  }

  project_class_map <- list()
  project_defaults_file <- cfg[["method_defaults_file"]]
  if (!is.null(project_defaults_file)) {
    project_class_map <- read_method_defaults_from_r_file(
      resolve_snapshot_defaults_file(project_defaults_file, root),
      source_name = "snapshot.method_defaults_file"
    )
  }

  inline_class_map <- coerce_class_override_map(
    cfg[["method_by_class"]],
    source = "snapshot.method_by_class"
  )

  merged_class_map <- pkg_class_map
  for (nm in names(project_class_map)) merged_class_map[[nm]] <- project_class_map[[nm]]
  for (nm in names(inline_class_map)) merged_class_map[[nm]] <- inline_class_map[[nm]]

  cfg[["method"]] <- if (is.null(cfg[["method"]])) {
    NULL
  } else {
    coerce_snapshot_methods_from_config(cfg[["method"]], arg_name = "snapshot.method")
  }
  cfg[["method_by_class"]] <- merged_class_map
  cfg
}

extract_method_labels_from_expr <- function(method_expr) {
  if (missing(method_expr) || is.null(method_expr)) {
    return(NULL)
  }
  if (is.symbol(method_expr)) {
    return(as.character(method_expr))
  }
  if (is.call(method_expr) && as.character(method_expr[[1L]]) %in% c("::", ":::")) {
    return(paste(deparse(method_expr), collapse = ""))
  }
  if (is.call(method_expr) && as.character(method_expr[[1L]]) %in% c("c", "list")) {
    args <- as.list(method_expr)[-1L]
    return(vapply(args, function(arg) paste(deparse(arg), collapse = ""), character(1L)))
  }
  NULL
}

coerce_snapshot_methods <- function(method,
                                    arg_name = "method",
                                    method_expr = NULL) {
  is_method_spec <- is.list(method) &&
    length(method) > 0L &&
    all(vapply(method, function(x) {
      is.list(x) &&
        is.function(x[["fn"]]) &&
        is.character(x[["label"]]) &&
        length(x[["label"]]) == 1L
    }, logical(1L)))
  if (is_method_spec) {
    return(method)
  }

  method_list <- if (is.function(method)) {
    list(method)
  } else if (is.list(method) && length(method) > 0L && all(vapply(method, is.function, logical(1L)))) {
    method
  } else {
    stop("`", arg_name, "` must be a function or a non-empty list of functions.", call. = FALSE)
  }

  labels <- names(method_list)
  if (is.null(labels)) {
    labels <- rep("", length(method_list))
  }

  expr_labels <- extract_method_labels_from_expr(method_expr)
  if (!is.null(expr_labels) && length(expr_labels) == length(method_list)) {
    missing_idx <- which(trimws(labels) == "")
    if (length(missing_idx) > 0L) {
      labels[missing_idx] <- expr_labels[missing_idx]
    }
  }

  guess_function_name <- function(fn) {
    envs <- list(
      globalenv(),
      baseenv(),
      asNamespace("base"),
      asNamespace("utils"),
      asNamespace("stats")
    )
    for (env in envs) {
      nms <- tryCatch(ls(env, all.names = TRUE), error = function(e) character())
      for (nm in nms) {
        obj <- tryCatch(get(nm, envir = env, inherits = FALSE), error = function(e) NULL)
        if (is.function(obj) && identical(obj, fn)) {
          return(nm)
        }
      }
    }
    NULL
  }

  guessed <- vapply(method_list, function(fn) {
    nm <- guess_function_name(fn)
    if (is.null(nm)) "" else nm
  }, character(1L))
  fill_guess <- which(trimws(labels) == "" & nzchar(guessed))
  if (length(fill_guess) > 0L) {
    labels[fill_guess] <- guessed[fill_guess]
  }

  missing_labels <- which(trimws(labels) == "")
  if (length(missing_labels) > 0L) {
    labels[missing_labels] <- paste0("unnamed_method_", missing_labels)
  }

  Map(function(fn, label) {
    list(fn = fn, label = label)
  }, method_list, labels)
}

resolve_snapshot_methods <- function(value, method, config, method_missing, method_expr = NULL) {
  if (!isTRUE(method_missing)) {
    return(coerce_snapshot_methods(method, arg_name = "method", method_expr = method_expr))
  }

  class_methods <- NULL
  cfg_snapshot <- config[["snapshot"]]
  by_class <- cfg_snapshot[["method_by_class"]]
  cls <- class(value)
  if (is.list(by_class) && length(cls) > 0L) {
    for (cl in cls) {
      if (!is.null(by_class[[cl]])) {
        class_methods <- by_class[[cl]]
        break
      }
    }
  }
  if (!is.null(class_methods)) return(class_methods)

  global_methods <- cfg_snapshot[["method"]]
  if (is.list(global_methods) && length(global_methods) > 0L) {
    return(global_methods)
  }
  coerce_snapshot_methods(DEFAULT_SNAPSHOT_METHODS, arg_name = "method")
}


#' Round Floating-Point Numbers in Snapshot Text
#'
#' Replaces every floating-point literal (decimal or scientific notation) found
#' in a character vector of snapshot lines with its value rounded to
#' \code{digits} decimal places.  Integer literals that contain no decimal
#' point and no exponent are left untouched, so index ranges such as
#' \code{[1:1071]} are never modified.
#'
#' \code{digits} is passed directly to \code{\link[base]{round}}: 0 rounds to
#' the nearest integer, negative values round to the left of the decimal point.
#'
#' @param text   Character vector of snapshot text lines.
#' @param digits Integer; number of decimal places to keep (passed to
#'   \code{round()}).  Must be a finite integer.
#'
#' @return Character vector with floating-point numbers rounded.
#'
#' @keywords internal
round_snapshot_numbers <- function(text, digits) {
  # Two alternatives, both require a decimal point or exponent so that pure
  # integers (e.g. index ranges like 1071) are never matched:
  #   1. A number WITH a decimal point, optionally followed by an exponent.
  #      Examples: 1.22   -0.423   2.41e-17   +1.0E+3
  #   2. An integer written in scientific notation (no decimal point required).
  #      Examples: 1e+07   -2E4
  pattern <- "[-+]?[0-9]*\\.[0-9]+([eE][-+]?[0-9]+)?|[-+]?[0-9]+[eE][-+]?[0-9]+"

  vapply(text, function(line) {
    m <- gregexpr(pattern, line, perl = TRUE)[[1]]
    if (m[1] == -1L) return(line)

    match_lengths <- attr(m, "match.length")
    parts <- character(length(m) * 2L + 1L)
    pos   <- 1L
    j     <- 1L

    for (i in seq_along(m)) {
      start <- m[i]
      end   <- start + match_lengths[i] - 1L

      parts[j] <- substr(line, pos, start - 1L)
      j <- j + 1L

      num_str <- substr(line, start, end)
      # The regex guarantees the matched string looks like a number; any failure
      # to parse (e.g. an NA) is treated as a non-match and the original token
      # is kept.  suppressWarnings avoids noisy NAs-introduced messages.
      num_val <- suppressWarnings(as.numeric(num_str))
      parts[j] <- if (!is.na(num_val)) {
        as.character(round(num_val, digits))
      } else {
        num_str
      }
      j   <- j + 1L
      pos <- end + 1L
    }

    parts[j] <- substr(line, pos, nchar(line))
    paste(parts[seq_len(j)], collapse = "")
  }, character(1L), USE.NAMES = FALSE)
}


#' Mask \code{[ignored]} Lines in New Snapshot Text
#'
#' Any line in \code{old_text} that equals \code{"[ignored]"} (after trimming
#' whitespace) causes the corresponding line in \code{new_text} to be replaced
#' with \code{"[ignored]"}, so that known-volatile lines never trigger a
#' snapshot failure.  Lines beyond the shorter vector are left unchanged.
#'
#' This helper is used both during comparison (so the lines are skipped) and
#' when writing an updated snapshot (so the markers are preserved).
#'
#' @param old_text Character vector of the stored snapshot lines.
#' @param new_text Character vector of the freshly generated snapshot lines.
#'
#' @return \code{new_text} with \code{[ignored]} substituted at matching
#'   positions.
#'
#' @keywords internal
mask_ignored_lines <- function(old_text, new_text) {
  min_len <- min(length(old_text), length(new_text))
  for (i in seq_len(min_len)) {
    if (trimws(old_text[i]) == IGNORED_MARKER) {
      new_text[i] <- IGNORED_MARKER
    }
  }
  new_text
}


# Fixed console width used when capturing snapshot output to ensure consistent,
# fully-untruncated output regardless of the R session's current width setting.
# The value is intentionally very large so that str() and print() never wrap
# or abbreviate values mid-line, keeping snapshot text deterministic across
# machines and environments.
SNAPSHOT_OUTPUT_WIDTH <- 110L

#' Serialize Value to Human-Readable Text
#'
#' Converts an R object to a human-readable text representation for snapshots.
#'
#' @param value The R object to serialize.
#' @param methods A function or a non-empty list of functions applied in order.
#'   Defaults to \code{list(print = base::print, str = utils::str)}.
#'
#' @return A character vector with the text representation.
#'
#' @keywords internal
serialize_value <- function(value, methods = NULL, use_class_defaults = TRUE) {
  if (is.null(methods) && use_class_defaults) {
    config <- tryCatch(
      read_resultcheck_config(),
      error = function(e) list(snapshot = list())
    )
    methods <- resolve_snapshot_methods(
      value = value,
      method = NULL,
      config = config,
      method_missing = TRUE
    )
  } else if (is.null(methods)) {
    methods <- DEFAULT_SNAPSHOT_METHODS
  }
  methods <- coerce_snapshot_methods(methods, arg_name = "methods")
  # Create a text representation using various methods
  output <- character()
  
  # Add header with object type
  output <- c(output, paste0("# Snapshot: ", class(value)[1]))
  output <- c(output, "")
  
  # Use a fixed large width so that snapshot output is consistent regardless
  # of the R session's console width setting.
  withr::with_options(list(width = SNAPSHOT_OUTPUT_WIDTH, pillar.advice = TRUE), {
    first_method <- TRUE
    for (method_def in methods) {
      section <- tryCatch(
        utils::capture.output(method_def$fn(value)),
        error = function(e) {
          obj_class <- class(value)
          if (length(obj_class) == 0L) obj_class <- as.character(typeof(value))
          stop(
            "Snapshot method `", method_def$label, "` is not available for class `",
            obj_class[1L], "`: ", conditionMessage(e),
            call. = FALSE
          )
        }
      )

      if (!first_method) {
        output <- c(output, "")
      }
      output <- c(output, paste0("## ", method_def$label))
      output <- c(output, section)
      first_method <- FALSE
    }
  })
  
  return(output)
}


#' Compare Two Snapshot Values
#'
#' Compares two serialized snapshots and returns differences.
#'
#' @param old_text  Character vector with old snapshot text.
#' @param new_text  Character vector with new snapshot text.
#' @param precision Optional integer.  When non-\code{NULL}, both texts are
#'   rounded to this many decimal places before comparison (see
#'   \code{\link{round_snapshot_numbers}}).  Useful for ignoring
#'   floating-point noise introduced by platform differences.
#'
#' @return A character vector of differences, or NULL if identical.
#'
#' @keywords internal
compare_snapshot_text <- function(old_text, new_text, precision = NULL) {
  old_text <- normalize_snapshot_text(old_text)
  new_text <- normalize_snapshot_text(new_text)

  # Apply numeric precision rounding when configured
  if (!is.null(precision)) {
    old_text <- round_snapshot_numbers(old_text, as.integer(precision))
    new_text <- round_snapshot_numbers(new_text, as.integer(precision))
  }

  # Mask lines that the user has marked as [ignored] in the stored snapshot
  new_text <- mask_ignored_lines(old_text, new_text)

  if (identical(old_text, new_text)) {
    return(NULL)
  }
  
  # Use waldo for comparison if available
  if (requireNamespace("waldo", quietly = TRUE)) {
    comparison <- waldo::compare(old_text, new_text, x_arg = "old", y_arg = "new")
    if (length(comparison) == 0) {
      return(NULL)
    }
    return(as.character(comparison))
  }
  
  # Fallback to basic diff
  output <- character()
  output <- c(output, "Snapshots differ:")
  
  # Show line-by-line differences
  max_len <- max(length(old_text), length(new_text))
  for (i in seq_len(max_len)) {
    old_line <- if (i <= length(old_text)) old_text[i] else "<missing>"
    new_line <- if (i <= length(new_text)) new_text[i] else "<missing>"
    
    if (!identical(old_line, new_line)) {
      output <- c(output, sprintf("Line %d:", i))
      output <- c(output, sprintf("  - %s", old_line))
      output <- c(output, sprintf("  + %s", new_line))
    }
  }
  
  return(output)
}


#' Normalize Snapshot Text Before Comparison
#'
#' Replaces volatile environment representations (for example memory addresses
#' in `.Environment` attributes) with a stable placeholder so snapshots remain
#' comparable across different execution contexts.
#'
#' @param text Character vector with serialized snapshot lines.
#'
#' @return Character vector with normalized snapshot lines.
#'
#' @keywords internal
normalize_snapshot_text <- function(text) {
  gsub(
    '(<environment: )[^>]+(>[[:space:]]*)$',
    '\\1<normalized>\\2',
    text
  )
}


#' Detect Testing Context
#'
#' Determines if code is running in a testing context (sandbox via run_in_sandbox).
#' This is used to change snapshot() behavior: interactive mode prompts for updates,
#' while testing mode throws errors on mismatch.
#'
#' @return Logical indicating if in testing mode.
#'
#' @keywords internal
is_testing <- function() {
  # Check if we're in a sandbox context by looking at call stack
  # run_in_sandbox stores a flag in the package environment
  in_sandbox <- !is.null(.resultcheck_env$.resultcheck_original_wd)
  
  return(in_sandbox)
}

warn_snapshot_write <- function(snapshot_file) {
  if (interactive()) {
    warning("snapshot() will write a snapshot file to: ", snapshot_file,
            call. = FALSE, immediate. = TRUE)
  }
}


#' Interactive Snapshot Testing
#'
#' Creates or updates a snapshot of an R object for interactive analysis.
#' On first use, saves the object to a human-readable snapshot file (.md).
#' On subsequent uses, compares the current object to the saved snapshot.
#'
#' In interactive mode (default), prompts the user to update if differences
#' are found and emits a warning. In testing mode (inside testthat or
#' run_in_sandbox), throws an error if snapshot doesn't exist or doesn't match.
#'
#' Snapshots are stored under \code{tests/_resultcheck_snaps/} by default,
#' organized by script name, and configurable via \code{snapshot.dir} in
#' \code{_resultcheck.yml}. Method defaults can be configured via
#' \code{snapshot.method} and class-specific defaults via
#' \code{snapshot.method_by_class}. Optional class defaults can also be loaded
#' from an R file using \code{snapshot.method_defaults_file}. Method strings in
#' config (for example \code{"print + str"} or \code{"stats::coef"}) are
#' resolved to callable functions. In config expressions, \code{"+"} is treated
#' as the method delimiter.
#'
#' Built-in class defaults (loaded from inst/extdata/snapshot-method-defaults.R)
#' use broom functions for many statistical model classes. The default method
#' is typically broom::tidy, with broom::glance and/or broom::augment
#' added where supported (per the broom available-methods table at
#' https://broom.tidymodels.org/articles/available-methods.html).
#'
#' @param value The R object to snapshot (e.g., plot, table, model output).
#' @param name Character. A descriptive name for this snapshot.
#' @param script_name Optional. The name of the script creating the snapshot.
#'   If NULL, auto-detects via \code{rstudioapi::getSourceEditorContext()$path}
#'   (in RStudio), falling back to the call stack, then to \code{"interactive"}.
#' @param method Optional function or non-empty list of functions used to
#'   serialize \code{value}. Functions are executed in order and each section
#'   header is taken from the method expression or list name. If omitted,
#'   method defaults are resolved in this order:
#'   \code{snapshot.method_by_class} (matched by object class), then
#'   \code{snapshot.method}, then
#'   \code{list(print = base::print, str = utils::str)}.
#'
#' @return Invisible TRUE if snapshot matches or was updated.
#'   In testing mode, throws an error if snapshot is missing or doesn't match.
#'
#' @export
#'
#' @examples
#' with_example({
#'   model <- stats::lm(mpg ~ wt, data = datasets::mtcars)
#'   snapshot(model, "model_default", script_name = "analysis")
#'   snapshot(model, "model_multi", script_name = "analysis",
#'            method = list(summary = summary, print = print))
#'   snapshot(model, "model_print", script_name = "analysis", method = print)
#'   snapshot(model, "model_ns", script_name = "analysis", method = stats::coef)
#'   snapshot(model, "model_length", script_name = "analysis", method = length)
#' })
#'
#' with_example({
#'   sandbox <- setup_sandbox()
#'   on.exit(cleanup_sandbox(sandbox), add = TRUE)
#'   run_in_sandbox("analysis.R", sandbox)
#' })
#'
#' if (interactive()) with_example({
#'   sandbox <- setup_sandbox()
#'   on.exit(cleanup_sandbox(sandbox), add = TRUE)
#'   run_in_sandbox("analysis.R", sandbox)
#' }, mismatch = TRUE)
snapshot <- function(value, name, script_name = NULL, method = NULL) {
  # Get snapshot file path (.md extension)
  snapshot_file <- get_snapshot_path(name, script_name, ext = "md")
  
  # Read project configuration for optional precision rounding
  config    <- read_resultcheck_config()
  precision <- config[["snapshot"]][["precision"]]
  methods <- resolve_snapshot_methods(
    value = value,
    method = method,
    config = config,
    method_missing = missing(method),
    method_expr = if (!missing(method)) substitute(method) else NULL
  )

  # Serialize the value to text
  new_text <- serialize_value(value, methods = methods)
  new_text <- normalize_snapshot_text(new_text)

  if (!is.null(precision)) {
    new_text <- round_snapshot_numbers(new_text, as.integer(precision))
  }
  
  # Detect if we're in testing mode
  testing_mode <- is_testing()
  
  # Check if snapshot exists
  if (!file.exists(snapshot_file)) {
    if (testing_mode) {
      # In testing mode, error if snapshot doesn't exist
      stop(
        "Snapshot does not exist: ", name, "\n",
        "File: ", snapshot_file, "\n",
        "Run the script interactively first to create snapshots.",
        call. = FALSE
      )
    }
    
    # First time: save the snapshot
    warn_snapshot_write(snapshot_file)
    writeLines(new_text, snapshot_file)
    message("\u2713 New snapshot saved: ", basename(dirname(snapshot_file)), "/", basename(snapshot_file))
    return(invisible(TRUE))
  }
  
  # Load existing snapshot
  old_text <- readLines(snapshot_file, warn = FALSE)
  
  # Compare snapshots (precision already applied to new_text above; also
  # applied to old_text inside compare_snapshot_text for backward compatibility
  # with snapshots stored before precision was configured).
  differences <- compare_snapshot_text(old_text, new_text, precision = precision)
  
  if (is.null(differences)) {
    # No differences - snapshot matches
    if (!testing_mode) {
      message("\u2713 Snapshot matches: ", name)
    }
    return(invisible(TRUE))
  }
  
  # Differences found
  diff_msg <- paste0(
    "\nSnapshot differences found for: ", name, "\n",
    "File: ", snapshot_file, "\n\n",
    "Differences:\n",
    paste(differences, collapse = "\n")
  )
  
  if (testing_mode) {
    # In testing mode, throw an error
    stop(diff_msg, "\n\nSnapshot does not match. Run interactively to review and update.", call. = FALSE)
  }
  
  # Interactive mode: show warning and prompt
  warning(diff_msg, call. = FALSE, immediate. = TRUE)
  
  if (interactive()) {
    # Prompt user to update
    cat("\n")
    response <- readline(prompt = "Update snapshot? (y/n): ")
    
    if (tolower(trimws(response)) == "y") {
      # Preserve any [ignored] markers from the stored snapshot
      warn_snapshot_write(snapshot_file)
      writeLines(mask_ignored_lines(old_text, new_text), snapshot_file)
      message("\u2713 Snapshot updated.")
      return(invisible(TRUE))
    } else {
      message("\u2717 Snapshot not updated.")
      return(invisible(FALSE))
    }
  } else {
    # Non-interactive but not testing - just warn
    message("\n\u26a0 Run interactively to update snapshot.")
    return(invisible(FALSE))
  }
}

test_that("snapshot works in sandbox (testing mode)", {
  skip_if_not_installed("withr")
  
  # Create a temporary project
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))
  
  # Create test data
  dir.create(file.path(temp_project, "data"))
  test_data <- data.frame(x = 1:10, y = rnorm(10, mean = 5, sd = 1))
  saveRDS(test_data, file.path(temp_project, "data", "input.rds"))
  
  # Create analysis script with snapshot
  script_content <- '
data <- readRDS("data/input.rds")
model <- lm(y ~ x, data = data)
snapshot(model, "test_model", script_name = "analysis")
saveRDS(model, "model.rds")
'
  script_path <- file.path(temp_project, "analysis.R")
  writeLines(script_content, script_path)
  
  on.exit(unlink(temp_project, recursive = TRUE), add = TRUE)
  
  # Change to project directory
  old_wd <- getwd()
  on.exit(setwd(old_wd), add = TRUE)
  setwd(temp_project)
  
  # Step 1: Run interactively to create snapshot
  data <- readRDS("data/input.rds")
  model <- lm(y ~ x, data = data)
  snapshot(model, "test_model", script_name = "analysis")
  
  snapshot_file <- file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "test_model.md")
  expect_true(file.exists(snapshot_file))
  expect_equal(tools::file_ext(snapshot_file), "md")
  
  # Step 2: Run in sandbox with same data (should pass)
  sandbox <- setup_sandbox(c("data/input.rds"))
  expect_silent(run_in_sandbox(script_path, sandbox))
  cleanup_sandbox(sandbox)
  
  # Step 3: Change data and run in sandbox (should error)
  different_data <- data.frame(x = 1:10, y = rnorm(10, mean = 8, sd = 1))
  saveRDS(different_data, file.path(temp_project, "data", "input.rds"))
  
  sandbox <- setup_sandbox(c("data/input.rds"))
  expect_error(
    run_in_sandbox(script_path, sandbox),
    "Snapshot differences found"
  )
  cleanup_sandbox(sandbox)
})

# These tests verify snapshot functionality
# Note: We test by sourcing scripts, not calling snapshot() directly in tests

test_that("find_root finds project root with .git", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))
  dir.create(file.path(temp_project, "subdir"), recursive = TRUE)
  
  on.exit(unlink(temp_project, recursive = TRUE))
  
  # Test from root directory
  withr::with_dir(temp_project, {
    root <- find_root()
    expect_equal(normalizePath(root), normalizePath(temp_project))
  })
  
  # Test from subdirectory
  withr::with_dir(file.path(temp_project, "subdir"), {
    root <- find_root()
    expect_equal(normalizePath(root), normalizePath(temp_project))
  })
})


test_that("snapshot creates .md file when run interactively", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))
  
  on.exit(unlink(temp_project, recursive = TRUE))
  
  withr::with_dir(temp_project, {
    test_value <- data.frame(x = 1:5, y = 6:10)
    snapshot(test_value, "test_snapshot", script_name = "analysis")
    
    snapshot_path <- file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "test_snapshot.md")
    expect_true(file.exists(snapshot_path))
    expect_equal(tools::file_ext(snapshot_path), "md")
    
    # Check content is human-readable
    content <- readLines(snapshot_path)
    expect_true(any(grepl("# Snapshot:", content)))
    expect_true(any(grepl("data.frame", content)))
  })
})


test_that("snapshot matches when run with same data", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))
  
  on.exit(unlink(temp_project, recursive = TRUE))
  
  withr::with_dir(temp_project, {
    test_value <- data.frame(x = 1:5, y = 6:10)
    
    # First run
    snapshot(test_value, "test_snapshot", script_name = "analysis")
    
    # Second run - should match
    expect_message(
      snapshot(test_value, "test_snapshot", script_name = "analysis"),
      "Snapshot matches"
    )
  })
})


test_that("snapshot works with different object types", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))
  
  on.exit(unlink(temp_project, recursive = TRUE))
  
  withr::with_dir(temp_project, {
    list_obj <- list(a = 1:5, b = "test")
    snapshot(list_obj, "list_snap", script_name = "analysis")
    
    model <- lm(mpg ~ wt, data = mtcars)
    snapshot(model, "model_snap", script_name = "analysis")
    
    vec <- c(1, 2, 3)
    snapshot(vec, "vec_snap", script_name = "analysis")

    model_snapshot <- readLines(
      file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "model_snap.md"),
      warn = FALSE
    )
    
    expect_true(file.exists(file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "list_snap.md")))
    expect_true(file.exists(file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "model_snap.md")))
    expect_true(file.exists(file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "vec_snap.md")))
    expect_true(any(grepl('<environment: <normalized>>', model_snapshot, fixed = TRUE)))
  })
})


test_that("snapshot organizes by script name when specified", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))
  
  on.exit(unlink(temp_project, recursive = TRUE))
  
  withr::with_dir(temp_project, {
    snapshot(data.frame(a = 1), "snap1", script_name = "custom1")
    snapshot(data.frame(b = 2), "snap2", script_name = "custom2")
    
    expect_true(dir.exists(file.path(temp_project, "tests/_resultcheck_snaps", "custom1")))
    expect_true(dir.exists(file.path(temp_project, "tests/_resultcheck_snaps", "custom2")))
  })
})


test_that("compare_snapshot_text ignores .Environment differences", {
  old_text <- c(
    "# Snapshot: lm",
    "",
    "## List Structure",
    "  .. ..- attr(*, \".Environment\")=<environment: 0x000001eb76e707a8> "
  )

  new_text <- c(
    "# Snapshot: lm",
    "",
    "## List Structure",
    "  .. ..- attr(*, \".Environment\")=<environment: R_GlobalEnv> "
  )

  differences <- resultcheck:::compare_snapshot_text(old_text, new_text)

  expect_null(differences)
})


test_that("serialize_value respects selected methods", {
  val <- list(a = 1:3, b = "hello")

  out_default <- resultcheck:::serialize_value(val)
  out_print <- resultcheck:::serialize_value(val, methods = print)
  out_str   <- resultcheck:::serialize_value(val, methods = str)
  out_multi <- resultcheck:::serialize_value(
    val,
    methods = list(summary = summary, print = print, summary_again = summary)
  )

  # "print" output contains the "## print" header but not "## str"
  expect_true(any(grepl("^## print$", out_print)))
  expect_false(any(grepl("^## str$", out_print)))
  expect_false(any(grepl("## List Structure", out_print)))

  # "str" output contains the "## str" header but not "## print"
  expect_true(any(grepl("^## str$", out_str)))
  expect_false(any(grepl("^## print$", out_str)))

  # default output includes print and str sections
  expect_true(any(grepl("^## print$", out_default)))
  expect_true(any(grepl("^## str$", out_default)))
  expect_false(any(grepl("## List Structure", out_default)))

  # methods are applied in provided order
  headers <- out_multi[grepl("^## ", out_multi)]
  expect_equal(headers, c("## summary", "## print", "## summary_again"))
})


test_that("snapshot respects method parameter", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))

  on.exit(unlink(temp_project, recursive = TRUE))

  withr::with_dir(temp_project, {
    val <- list(a = 1:3, id = "volatile_abc123")

    snapshot(val, "snap_print", script_name = "analysis", method = print)
    snapshot(val, "snap_str",   script_name = "analysis", method = str)
    snapshot(
      val,
      "snap_expr",
      script_name = "analysis",
      method = list(print = print, summary = summary)
    )
    snapshot(val, "snap_length", script_name = "analysis", method = length)

    content_print <- readLines(
      file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "snap_print.md"),
      warn = FALSE
    )
    content_str <- readLines(
      file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "snap_str.md"),
      warn = FALSE
    )
    content_expr <- readLines(
      file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "snap_expr.md"),
      warn = FALSE
    )
    content_length <- readLines(
      file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "snap_length.md"),
      warn = FALSE
    )

    # print-only snapshot has "## print" header
    expect_true(any(grepl("^## print$", content_print)))
    # str-only snapshot has "## str" header
    expect_true(any(grepl("^## str$", content_str)))

    # Neither should contain headers belonging to the other method
    expect_false(any(grepl("^## str$", content_print)))
    expect_false(any(grepl("^## print$", content_str)))

    expect_true(any(grepl("^## print$", content_expr)))
    expect_true(any(grepl("^## summary$", content_expr)))

    expect_true(any(grepl("^## length$", content_length)))
  })
})

test_that("coerce_snapshot_methods validates function-based methods", {
  expect_error(
    resultcheck:::coerce_snapshot_methods("print"),
    "must be a function or a non-empty list of functions"
  )
  expect_error(
    resultcheck:::coerce_snapshot_methods(list(print, "str")),
    "must be a function or a non-empty list of functions"
  )

  methods <- resultcheck:::coerce_snapshot_methods(list(print = print, str = str))
  expect_equal(unname(vapply(methods, `[[`, character(1), "label")), c("print", "str"))
  expect_identical(methods[[1]]$fn, print)
  expect_identical(methods[[2]]$fn, str)

  guessed <- resultcheck:::coerce_snapshot_methods(list(print, str))
  expect_equal(unname(vapply(guessed, `[[`, character(1), "label")), c("print", "str"))

  namespaced <- resultcheck:::coerce_snapshot_methods(
    stats::coef,
    method_expr = substitute(stats::coef)
  )
  expect_equal(unname(vapply(namespaced, `[[`, character(1), "label")), "stats::coef")

  fallback <- resultcheck:::coerce_snapshot_methods(list(function(x) x))
  expect_equal(unname(vapply(fallback, `[[`, character(1), "label")), "unnamed_method_1")
})

test_that("serialize_value reports method execution failures clearly", {
  assign("summary.fail_summary", function(object, ...) {
    stop("summary failed intentionally")
  }, envir = .GlobalEnv)
  on.exit(rm("summary.fail_summary", envir = .GlobalEnv), add = TRUE)

  x <- structure(list(a = 1), class = "fail_summary")
  expect_error(
    resultcheck:::serialize_value(x, methods = summary),
    "Snapshot method `summary` is not available for class `fail_summary`: summary failed intentionally"
  )
})


# ── [ignored] marker tests ────────────────────────────────────────────────────

test_that("mask_ignored_lines replaces new lines at [ignored] positions", {
  old_text <- c("line 1", "[ignored]", "line 3")
  new_text <- c("line 1", "different value", "line 3")

  result <- resultcheck:::mask_ignored_lines(old_text, new_text)
  expect_equal(result, c("line 1", "[ignored]", "line 3"))
})


test_that("mask_ignored_lines handles [ignored] with surrounding whitespace", {
  old_text <- c("  [ignored]  ", "normal line")
  new_text <- c("some value",    "normal line")

  result <- resultcheck:::mask_ignored_lines(old_text, new_text)
  expect_equal(result, c("[ignored]", "normal line"))
})


test_that("compare_snapshot_text returns NULL when only [ignored] lines differ", {
  old_text <- c(
    "# Snapshot: numeric",
    "",
    "## Value",
    "[ignored]",
    "[ignored]"
  )
  new_text <- c(
    "# Snapshot: numeric",
    "",
    "## Value",
    "  ..- attr(*, \"scaled:center\")= num 2.41e-17",
    "  ..- attr(*, \"scaled:center\")= num 2.98e-17"
  )

  result <- resultcheck:::compare_snapshot_text(old_text, new_text)
  expect_null(result)
})


test_that("snapshot in testing mode passes when differing lines are [ignored]", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))

  on.exit(unlink(temp_project, recursive = TRUE))

  withr::with_dir(temp_project, {
    val1 <- c(1.0, 2.0, 3.0)
    snapshot(val1, "snap_ignored_mode", script_name = "analysis")

    snap_file <- file.path(temp_project, "tests/_resultcheck_snaps",
                           "analysis", "snap_ignored_mode.md")

    lines <- readLines(snap_file)
    # Match changed numeric output in both print() ("[1] ...") and str() ("num [...] ...")
    numeric_output_pattern <- "^\\[1\\]|^\\s*num \\["
    value_lines <- which(grepl(numeric_output_pattern, lines))
    lines[value_lines] <- "[ignored]"
    writeLines(lines, snap_file)

    # Simulate testing mode by writing a script and running it in a sandbox
    script_content <- paste0(
      'val2 <- c(99.0, 2.0, 3.0)\n',
      'resultcheck::snapshot(val2, "snap_ignored_mode", script_name = "analysis")\n'
    )
    writeLines(script_content, file.path(temp_project, "test_script.R"))

    sandbox <- setup_sandbox(character(0))
    expect_no_error(run_in_sandbox("test_script.R", sandbox))
    cleanup_sandbox(sandbox)
  })
})


test_that("mask_ignored_lines preserves [ignored] marker in masked new_text", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))

  on.exit(unlink(temp_project, recursive = TRUE))

  withr::with_dir(temp_project, {
    val <- c(1.0, 2.0, 3.0)
    snapshot(val, "ignored_preserve_test", script_name = "analysis")

    snap_file <- file.path(temp_project, "tests/_resultcheck_snaps",
                           "analysis", "ignored_preserve_test.md")

    lines <- readLines(snap_file)
    value_line <- which(grepl("^\\[1\\]", lines))[1]
    lines[value_line] <- "[ignored]"
    writeLines(lines, snap_file)

    old_text <- readLines(snap_file)
    val2 <- c(99.0, 2.0, 3.0)
    new_text <- resultcheck:::serialize_value(val2, methods = list(print = print, str = str))
    new_text <- resultcheck:::normalize_snapshot_text(new_text)

    masked <- resultcheck:::mask_ignored_lines(old_text, new_text)
    expect_equal(masked[value_line], "[ignored]")
  })
})


# ── precision rounding tests ──────────────────────────────────────────────────

test_that("round_snapshot_numbers rounds decimal numbers", {
  text <- c("num 2.41e-17", "num 1.22", "int 5")

  result <- resultcheck:::round_snapshot_numbers(text, digits = 10L)

  expect_equal(result[1], "num 0")
  expect_equal(result[2], "num 1.22")
  expect_equal(result[3], "int 5")
})


test_that("round_snapshot_numbers leaves non-numeric text unchanged", {
  text <- c("# Snapshot: data.frame", "## Structure", "chr \"hello\"")
  result <- resultcheck:::round_snapshot_numbers(text, digits = 5L)
  expect_equal(result, text)
})


test_that("compare_snapshot_text returns NULL with precision when tiny diffs exist", {
  old_text <- c(
    "# Snapshot: numeric",
    "",
    "## Value",
    "  ..- attr(*, \"scaled:center\")= num 2.41e-17"
  )
  new_text <- c(
    "# Snapshot: numeric",
    "",
    "## Value",
    "  ..- attr(*, \"scaled:center\")= num 1.63e-17"
  )

  expect_false(is.null(
    resultcheck:::compare_snapshot_text(old_text, new_text)
  ))
  expect_null(
    resultcheck:::compare_snapshot_text(old_text, new_text, precision = 10L)
  )
})


test_that("snapshot reads precision from _resultcheck.yml and stores rounded values", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))

  on.exit(unlink(temp_project, recursive = TRUE))

  writeLines(c("snapshot:", "  precision: 5"),
             file.path(temp_project, "_resultcheck.yml"))

  withr::with_dir(temp_project, {
    val <- 1.123456789
    snapshot(val, "precision_test", script_name = "analysis")

    snap_file <- file.path(temp_project, "tests/_resultcheck_snaps",
                           "analysis", "precision_test.md")
    content <- readLines(snap_file)
    expect_true(any(grepl("1.12346", content, fixed = TRUE)))
    expect_false(any(grepl("1.123456789", content, fixed = TRUE)))
  })
})

test_that("snapshot uses configured snapshot.dir from _resultcheck.yml", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))

  on.exit(unlink(temp_project, recursive = TRUE))

  writeLines(
    c("snapshot:", "  dir: custom_snapshots"),
    file.path(temp_project, "_resultcheck.yml")
  )

  withr::with_dir(temp_project, {
    snapshot(letters[1:3], "custom_dir_test", script_name = "analysis")

    snap_file <- file.path(
      temp_project, "custom_snapshots", "analysis", "custom_dir_test.md"
    )
    expect_true(file.exists(snap_file))
  })
})

test_that("read_resultcheck_config supports legacy resultcheck.yml", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))

  on.exit(unlink(temp_project, recursive = TRUE))

  writeLines(
    c("snapshot:", "  precision: 4"),
    file.path(temp_project, "resultcheck.yml")
  )

  withr::with_dir(temp_project, {
    cfg <- resultcheck:::read_resultcheck_config()
    expect_equal(cfg$snapshot$precision, 4)
  })
})


test_that("read_resultcheck_config returns empty list when no yml exists", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))

  on.exit(unlink(temp_project, recursive = TRUE))

  withr::with_dir(temp_project, {
    cfg <- resultcheck:::read_resultcheck_config()
    expect_type(cfg, "list")
    expect_length(cfg, 0L)
  })
})

test_that("snapshot defaults to print and str when method is omitted", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))
  on.exit(unlink(temp_project, recursive = TRUE))

  withr::with_dir(temp_project, {
    # Use a plain named list which has no built-in class default
    val <- list(a = 1L, b = "hello")
    snapshot(val, "default_methods_test", script_name = "analysis")

    content <- readLines(
      file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "default_methods_test.md"),
      warn = FALSE
    )
    expect_true(any(grepl("^## print$", content)))
    expect_true(any(grepl("^## str$", content)))
  })
})

test_that("snapshot rejects character method values", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))
  on.exit(unlink(temp_project, recursive = TRUE))

  withr::with_dir(temp_project, {
    model <- lm(mpg ~ wt, data = mtcars)
    expect_error(
      snapshot(model, "char_method_test", script_name = "analysis", method = "print"),
      "must be a function or a non-empty list of functions"
    )
  })
})

test_that("snapshot uses class default methods from _resultcheck.yml", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))
  on.exit(unlink(temp_project, recursive = TRUE))

  writeLines(
    c(
      "snapshot:",
      "  method_by_class:",
      "    lm: summary"
    ),
    file.path(temp_project, "_resultcheck.yml")
  )

  withr::with_dir(temp_project, {
    model <- lm(mpg ~ wt, data = mtcars)
    snapshot(model, "class_default_test", script_name = "analysis")
    content <- readLines(
      file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "class_default_test.md"),
      warn = FALSE
    )
    expect_true(any(grepl("^## summary$", content)))
    expect_false(any(grepl("^## print$", content)))
  })
})

test_that("explicit method overrides class default methods", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))
  on.exit(unlink(temp_project, recursive = TRUE))

  writeLines(
    c(
      "snapshot:",
      "  method_by_class:",
      "    lm: summary"
    ),
    file.path(temp_project, "_resultcheck.yml")
  )

  withr::with_dir(temp_project, {
    model <- lm(mpg ~ wt, data = mtcars)
    snapshot(model, "class_override_test", script_name = "analysis", method = print)
    content <- readLines(
      file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "class_override_test.md"),
      warn = FALSE
    )
    expect_true(any(grepl("^## print$", content)))
    expect_false(any(grepl("^## summary$", content)))
  })
})

test_that("snapshot can read class defaults from separate R defaults file", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))
  on.exit(unlink(temp_project, recursive = TRUE))

  defaults_path <- file.path(temp_project, "snapshot-method-overrides.R")
  writeLines(
    "method_by_class <- list(lm = \"summary\")",
    defaults_path
  )
  writeLines(
    c(
      "snapshot:",
      "  method_defaults_file: snapshot-method-overrides.R"
    ),
    file.path(temp_project, "_resultcheck.yml")
  )

  withr::with_dir(temp_project, {
    cfg <- resultcheck:::read_resultcheck_config()
    expect_equal(cfg$snapshot$method_by_class$lm[[1]]$label, "summary")

    model <- lm(mpg ~ wt, data = mtcars)
    snapshot(model, "class_file_default_test", script_name = "analysis")
    content <- readLines(
      file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "class_file_default_test.md"),
      warn = FALSE
    )
    expect_true(any(grepl("^## summary$", content)))
    expect_false(any(grepl("^## print$", content)))
  })
})

test_that("inline class overrides take precedence over defaults file", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))
  on.exit(unlink(temp_project, recursive = TRUE))

  defaults_path <- file.path(temp_project, "snapshot-method-overrides.R")
  writeLines(
    "method_by_class <- list(lm = \"print\")",
    defaults_path
  )
  writeLines(
    c(
      "snapshot:",
      "  method_defaults_file: snapshot-method-overrides.R",
      "  method_by_class:",
      "    lm: summary"
    ),
    file.path(temp_project, "_resultcheck.yml")
  )

  withr::with_dir(temp_project, {
    cfg <- resultcheck:::read_resultcheck_config()
    expect_equal(cfg$snapshot$method_by_class$lm[[1]]$label, "summary")
  })
})

test_that("snapshot uses global method default when method is omitted", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))
  on.exit(unlink(temp_project, recursive = TRUE))

  writeLines(
    c(
      "snapshot:",
      "  method: summary + print"
    ),
    file.path(temp_project, "_resultcheck.yml")
  )

  withr::with_dir(temp_project, {
    # Use a numeric vector (no class default) - global method is used
    vec <- c(1, 2, 3)
    snapshot(vec, "global_default_test", script_name = "analysis")
    content <- readLines(
      file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "global_default_test.md"),
      warn = FALSE
    )
    headers <- content[grepl("^## ", content)]
    # Global method is used when no class default exists
    expect_equal(headers, c("## summary", "## print"))
    expect_true(any(grepl("Min.", content)))
  })
})

test_that("snapshot labels namespaced methods from expression", {
  temp_project <- tempfile()
  dir.create(temp_project)
  dir.create(file.path(temp_project, ".git"))
  on.exit(unlink(temp_project, recursive = TRUE))

  withr::with_dir(temp_project, {
    model <- lm(mpg ~ wt, data = mtcars)
    snapshot(model, "namespaced_method_test", script_name = "analysis", method = stats::coef)
    content <- readLines(
      file.path(temp_project, "tests/_resultcheck_snaps", "analysis", "namespaced_method_test.md"),
      warn = FALSE
    )
    expect_true(any(grepl("^## stats::coef$", content)))
  })
})

#' Get project root for sandbox operations
#' 
#' Internal helper to get project root, falling back to current directory
#' if root cannot be determined.
#' 
#' @return Character path to project root or current working directory
#' @keywords internal
.get_project_root <- function() {
  tryCatch({
    find_root()
  }, error = function(e) {
    getwd()
  })
}


#' Run Code Inside a Temporary Example Project
#'
#' Creates a self-contained example project under \code{tempdir()}, including:
#' \itemize{
#'   \item \code{_resultcheck.yml} (project root marker)
#'   \item \code{analysis.R} with \code{snapshot(model, "model")}
#'   \item matching and mismatched snapshot files
#'   \item \code{tests/testthat/test-analysis.R}
#' }
#' then temporarily sets the working directory to that project while
#' evaluating \code{code}.
#'
#' @param code Code to evaluate inside the temporary example project.
#' @param mismatch Logical. If TRUE, replaces the active snapshot with a
#'   mismatched version before evaluating \code{code}.
#'
#' @return The value of \code{code}.
#'
#' @export
#'
#' @examples
#' with_example({
#'   root <- find_root()
#'   print(root)
#' })
with_example <- function(code, mismatch = FALSE) {
  example_root <- tempfile("resultcheck-example-")
  dir.create(example_root, recursive = TRUE, showWarnings = FALSE)
  on.exit(unlink(example_root, recursive = TRUE, force = TRUE), add = TRUE)

  dir.create(file.path(example_root, "tests", "_resultcheck_snaps", "analysis"),
             recursive = TRUE, showWarnings = FALSE)
  dir.create(file.path(example_root, "tests", "testthat"),
             recursive = TRUE, showWarnings = FALSE)

  writeLines("# Example project root marker",
             file.path(example_root, "_resultcheck.yml"))

  writeLines(
    c(
      'model <- stats::lm(mpg ~ wt, data = datasets::mtcars)',
      'resultcheck::snapshot(model, "model")'
    ),
    file.path(example_root, "analysis.R")
  )

  # Create reference snapshots from inside the example project so that
  # find_root() resolves to example_root and serialize_value() picks up the
  # same built-in method defaults (e.g. broom for lm) that snapshot() will
  # see when run inside the sandbox.
  snapshot_path <- file.path(
    example_root, "tests", "_resultcheck_snaps", "analysis", "model.md"
  )
  mismatch_path <- file.path(
    example_root, "tests", "_resultcheck_snaps", "analysis", "model_mismatch.md"
  )

  withr::with_dir(example_root, {
    model <- stats::lm(mpg ~ wt, data = datasets::mtcars)
    matching_snapshot_text <- serialize_value(model)
    writeLines(matching_snapshot_text, snapshot_path)

    mismatch_text <- matching_snapshot_text
    # Intentionally alter the first numeric token so the mismatch demo fails.
    numeric_pattern <- "[-+]?[0-9]*\\.?[0-9]+([eE][-+]?[0-9]+)?"
    first_numeric_line <- which(grepl(numeric_pattern, mismatch_text, perl = TRUE))[1]
    if (!is.na(first_numeric_line)) {
      line <- mismatch_text[first_numeric_line]
      m <- regexpr(numeric_pattern, line, perl = TRUE)
      if (m[1] != -1L) {
        matched <- regmatches(line, m)
        digits <- if (grepl("\\.", matched)) nchar(sub(".*\\.", "", matched)) else 0L
        altered_number <- format(
          round(as.numeric(matched) + 1, digits),
          nsmall = digits,
          scientific = FALSE
        )
        regmatches(line, m) <- altered_number
        mismatch_text[first_numeric_line] <- line
      }
    }
    writeLines(mismatch_text, mismatch_path)
  })

  writeLines(
    c(
      "library(testthat)",
      "library(resultcheck)",
      "",
      'test_that("analysis produces stable results", {',
      '  sandbox <- setup_sandbox()',
      "  on.exit(cleanup_sandbox(sandbox), add = TRUE)",
      "",
      '  expect_true(run_in_sandbox("analysis.R", sandbox))',
      "})"
    ),
    file.path(example_root, "tests", "testthat", "test-analysis.R")
  )

  if (isTRUE(mismatch)) {
    file.copy(mismatch_path, snapshot_path, overwrite = TRUE)
  }

  if (!requireNamespace("withr", quietly = TRUE)) {
    stop("Package 'withr' is required but not installed. ",
         "Please install it with: install.packages('withr')")
  }

  expr <- substitute(code)
  withr::with_dir(example_root, eval.parent(expr))
}


#' Setup a Sandbox Environment for Testing
#'
#' Creates a temporary directory and copies specified files and/or directories
#' into it while preserving their path structure. This is useful for testing
#' empirical analysis scripts in isolation.
#'
#' @param files Character vector of relative file or directory paths to copy
#'   to the sandbox. Leave as \code{NULL} (default) to create an empty sandbox.
#'   Paths are resolved relative to the project root (found
#'   using \code{find_root()}); if the project root cannot be determined the
#'   current working directory is used.  When a path refers to a directory,
#'   the entire directory is copied recursively.  Absolute paths and path
#'   traversal attempts (e.g., \code{..}) are rejected for security.
#'   Snapshot files do \emph{not} need to be listed here: \code{snapshot()}
#'   always reads snapshots from the project root, not from the sandbox.
#' @param temp_base Optional. Custom location for the temporary directory.
#'   If NULL (default), uses \code{tempfile()}.
#'
#' @return A list with class "resultcheck_sandbox" containing:
#'   \item{path}{The path to the created temporary directory}
#'   \item{id}{A unique timestamp-based identifier for this sandbox}
#'
#' @export
#'
#' @examples
#' with_example({
#'   sandbox <- setup_sandbox()
#'   print(sandbox$path)
#'   cleanup_sandbox(sandbox)
#' })
setup_sandbox <- function(files = NULL, temp_base = NULL) {
  if (is.null(files)) {
    files <- character(0)
  }
  if (!is.character(files)) {
    stop("`files` must be a character vector of relative paths or NULL.", call. = FALSE)
  }

  # Generate unique ID for this sandbox
  sandbox_id <- paste0(
    "sandbox_",
    format(Sys.time(), "%Y%m%d_%H%M%S"),
    "_",
    paste(sample(c(letters, 0:9), 8, replace = TRUE), collapse = "")
  )
  
  # Create temp directory
  if (is.null(temp_base)) {
    temp_dir <- tempfile(pattern = sandbox_id)
  } else {
    temp_dir <- file.path(temp_base, sandbox_id)
  }
  
  # Try to create the directory
  tryCatch({
    dir.create(temp_dir, recursive = TRUE, showWarnings = FALSE)
    if (!dir.exists(temp_dir)) {
      stop("Failed to create sandbox directory: ", temp_dir)
    }
  }, error = function(e) {
    stop("Failed to create sandbox directory: ", e$message)
  })
  
  # Find project root to resolve file paths relative to it
  project_root <- .get_project_root()
  
  # Copy files/directories while preserving path structure
  for (file in files) {
    # Validate that path is relative (no absolute paths allowed)
    if (.Platform$OS.type == "windows") {
      # On Windows, check for drive letters or UNC paths
      is_absolute <- grepl("^[A-Za-z]:|^\\\\\\\\|^/", file)
    } else {
      # On Unix-like systems, check for leading slash
      is_absolute <- grepl("^/", file)
    }
    
    if (is_absolute) {
      stop("Absolute paths are not allowed. Please use relative paths only: ", file)
    }
    
    # Validate that path doesn't contain path traversal attempts
    # Split path and check each component for exactly ".."
    path_components <- strsplit(file, "[/\\\\]")[[1]]
    if (any(path_components == "..")) {
      stop("Path traversal (e.g., '..') is not allowed for security reasons: ", file)
    }
    
    # Resolve path relative to project root
    full_file_path <- file.path(project_root, file)
    
    if (dir.exists(full_file_path)) {
      # Copy directory recursively, preserving the relative path structure
      all_sub_files <- list.files(full_file_path, recursive = TRUE,
                                  all.files = TRUE, full.names = FALSE)
      for (subfile in all_sub_files) {
        src <- file.path(full_file_path, subfile)
        if (dir.exists(src)) next  # skip directory entries
        dst <- file.path(temp_dir, file, subfile)
        if (!dir.exists(dirname(dst))) {
          dir.create(dirname(dst), recursive = TRUE, showWarnings = FALSE)
        }
        tryCatch({
          file.copy(src, dst, overwrite = TRUE)
        }, error = function(e) {
          warning("Failed to copy file ", file.path(file, subfile), ": ", e$message)
        })
      }
    } else if (file.exists(full_file_path)) {
      # Determine the target path
      target_path <- file.path(temp_dir, file)
      
      # Create parent directories if needed
      target_dir <- dirname(target_path)
      if (!dir.exists(target_dir)) {
        dir.create(target_dir, recursive = TRUE, showWarnings = FALSE)
      }
      
      # Copy the file
      # Note: file.copy() follows symlinks and copies the target file's content,
      # not the symlink itself. This means even if 'file' is a symlink pointing
      # outside the current directory, the copied content will be in the sandbox.
      # This is the desired behavior for creating isolated test environments.
      tryCatch({
        file.copy(full_file_path, target_path, overwrite = TRUE)
      }, error = function(e) {
        warning("Failed to copy file ", file, ": ", e$message)
      })
    } else {
      warning("File not found, skipping: ", file)
    }
  }
  
  # Create and return sandbox object
  sandbox <- list(
    path = temp_dir,
    id = sandbox_id
  )
  class(sandbox) <- c("resultcheck_sandbox", "list")
  
  # Store in package environment for later retrieval
  .resultcheck_env$last_sandbox <- sandbox
  
  return(sandbox)
}


#' Run Code in a Sandbox Environment
#'
#' Executes an R script within a sandbox directory, suppressing messages,
#' warnings, and graphical output. This is useful for testing empirical
#' analysis scripts without polluting the console or creating unwanted plots.
#'
#' @param script_path Path to the R script to execute.
#' @param sandbox Optional. A sandbox object created by \code{setup_sandbox()}.
#'   If NULL (default), uses the most recently created sandbox.
#' @param suppress_messages Logical. Whether to suppress messages (default: TRUE).
#' @param suppress_warnings Logical. Whether to suppress warnings (default: TRUE).
#' @param capture_output Logical. Whether to capture output (default: TRUE).
#'
#' @return Invisible TRUE on successful execution.
#'
#' @export
#'
#' @examples
#' with_example({
#'   sandbox <- setup_sandbox()
#'   on.exit(cleanup_sandbox(sandbox), add = TRUE)
#'   run_in_sandbox("analysis.R", sandbox)
#' })
run_in_sandbox <- function(script_path, 
                           sandbox = NULL, 
                           suppress_messages = TRUE,
                           suppress_warnings = TRUE,
                           capture_output = TRUE) {
  # Get sandbox
  if (is.null(sandbox)) {
    if (is.null(.resultcheck_env$last_sandbox)) {
      stop("No sandbox specified and no previous sandbox found. ",
           "Please create a sandbox with setup_sandbox() first.")
    }
    sandbox <- .resultcheck_env$last_sandbox
  }
  
  # Validate sandbox
  if (!inherits(sandbox, "resultcheck_sandbox")) {
    stop("Invalid sandbox object. Please provide a sandbox created by setup_sandbox().")
  }
  
  if (!dir.exists(sandbox$path)) {
    stop("Sandbox directory does not exist: ", sandbox$path)
  }
  
  # Check if withr is available
  if (!requireNamespace("withr", quietly = TRUE)) {
    stop("Package 'withr' is required but not installed. ",
         "Please install it with: install.packages('withr')")
  }
  
  # Check if script exists in sandbox (preferred) or in project root
  script_in_sandbox <- file.path(sandbox$path, script_path)
  if (file.exists(script_in_sandbox)) {
    # Script was copied to sandbox, use it from there
    script_to_run <- script_path
  } else {
    # Try to find script in project root
    project_root <- .get_project_root()
    script_in_root <- file.path(project_root, script_path)
    
    if (file.exists(script_in_root)) {
      # Copy script to sandbox before running, preserving directory structure
      target_path <- file.path(sandbox$path, script_path)
      target_dir <- dirname(target_path)
      if (!dir.exists(target_dir)) {
        dir.create(target_dir, recursive = TRUE, showWarnings = FALSE)
      }
      file.copy(script_in_root, target_path, overwrite = TRUE)
      script_to_run <- script_path
    } else if (file.exists(script_path)) {
      # Fallback: script exists at an absolute path or relative to current directory
      # Check if it's an absolute path
      is_absolute <- if (.Platform$OS.type == "windows") {
        grepl("^[A-Za-z]:|^\\\\\\\\|^/", script_path)
      } else {
        grepl("^/", script_path)
      }
      
      if (is_absolute) {
        # For absolute paths, copy to sandbox root with basename only
        target_path <- file.path(sandbox$path, basename(script_path))
        file.copy(script_path, target_path, overwrite = TRUE)
        script_to_run <- basename(script_path)
      } else {
        # For relative paths, preserve directory structure
        target_path <- file.path(sandbox$path, script_path)
        target_dir <- dirname(target_path)
        if (!dir.exists(target_dir)) {
          dir.create(target_dir, recursive = TRUE, showWarnings = FALSE)
        }
        file.copy(script_path, target_path, overwrite = TRUE)
        script_to_run <- script_path
      }
    } else {
      stop("Script file not found: ", script_path)
    }
  }
  
  # Build the execution expression
  exec_expr <- quote(source(script_to_run, keep.source = TRUE))
  
  if (capture_output) {
    exec_expr <- bquote(capture.output(.(exec_expr)))
  }
  
  if (suppress_warnings) {
    exec_expr <- bquote(suppressWarnings(.(exec_expr)))
  }
  
  if (suppress_messages) {
    exec_expr <- bquote(suppressMessages(.(exec_expr)))
  }
  
  # Store original working directory for snapshot functions
  original_wd <- getwd()
  
  # Execute in sandbox directory with graphics suppressed
  tryCatch({
    withr::with_dir(sandbox$path, {
      # Store original WD in package environment so snapshot() can find project root
      .resultcheck_env$.resultcheck_original_wd <- original_wd
      on.exit({
        .resultcheck_env$.resultcheck_original_wd <- NULL
      }, add = TRUE)
      
      pdf(NULL)
      eval(exec_expr)
      dev.off()
    })
  }, error = function(e) {
    # Ensure cleanup happens even on error
    .resultcheck_env$.resultcheck_original_wd <- NULL
    stop("Error executing script in sandbox: ", e$message)
  })
  
  # Ensure cleanup (redundant but safe)
  .resultcheck_env$.resultcheck_original_wd <- NULL
  
  invisible(TRUE)
}


#' Clean Up a Sandbox Environment
#'
#' Removes a sandbox directory and all its contents. This should be called
#' after testing is complete to free up disk space.
#'
#' @param sandbox Optional. A sandbox object created by \code{setup_sandbox()}.
#'   If NULL (default), cleans up the most recently created sandbox.
#' @param force Logical. If TRUE (default), removes directory even if it
#'   contains files.
#'
#' @return Logical indicating success (invisible).
#'
#' @export
#'
#' @examples
#' with_example({
#'   sandbox <- setup_sandbox()
#'   cleanup_sandbox(sandbox)
#' })
cleanup_sandbox <- function(sandbox = NULL, force = TRUE) {
  # Get sandbox
  if (is.null(sandbox)) {
    if (is.null(.resultcheck_env$last_sandbox)) {
      warning("No sandbox specified and no previous sandbox found. Nothing to clean up.")
      return(invisible(FALSE))
    }
    sandbox <- .resultcheck_env$last_sandbox
  }
  
  # Validate sandbox
  if (!inherits(sandbox, "resultcheck_sandbox")) {
    stop("Invalid sandbox object. Please provide a sandbox created by setup_sandbox().")
  }
  
  temp_dir <- sandbox$path
  
  # Check if directory exists
  if (!dir.exists(temp_dir)) {
    warning("Sandbox directory does not exist: ", temp_dir)
    return(invisible(FALSE))
  }
  
  # Remove the directory
  tryCatch({
    unlink(temp_dir, recursive = TRUE, force = force)
    
    # Clear from package environment if it's the last sandbox
    if (!is.null(.resultcheck_env$last_sandbox) && 
        identical(.resultcheck_env$last_sandbox$path, temp_dir)) {
      .resultcheck_env$last_sandbox <- NULL
    }
    
    return(invisible(TRUE))
  }, error = function(e) {
    warning("Failed to clean up sandbox directory: ", e$message)
    return(invisible(FALSE))
  })
}


# Package environment to store state
.resultcheck_env <- new.env(parent = emptyenv())
.resultcheck_env$last_sandbox <- NULL

# Built-in class-based snapshot method defaults.
# Define `method_by_class` as a named list where each value is a method
# expression string (for example "summary") or a method spec.
#
# These defaults provide sensible snapshotting methods for common R classes:
# - "print"  : uses the object's print method
# - "str"    : uses str() to show structure
# - "summary": uses summary() (best for statistical models like lm, glm)
# - etc. (any function accessible via name or pkg::name syntax)
method_by_class <- list(
  # Linear models - broom::tidy gives useful coefficients table
  lm         = c("broom::tidy", "broom::glance", "broom::augment"),
  glm        = c("broom::tidy", "broom::glance", "broom::augment"),
  aov        = c("broom::tidy", "broom::glance"),
  manova     = "broom::tidy",

  # Extended linear models (broom supports tidy + glance for all)
  lm.beta    = "broom::tidy",
  lmodel2    = c("broom::tidy", "broom::glance"),
  biglm      = c("broom::tidy", "broom::glance"),
  speedlm    = c("broom::tidy", "broom::glance", "broom::augment"),

  # Robust linear models
  rlm        = c("broom::tidy", "broom::glance", "broom::augment"),
  lmrob      = c("broom::tidy", "broom::glance", "broom::augment"),
  lmRob      = c("broom::tidy", "broom::glance", "broom::augment"),
  glmrob     = "broom::tidy",
  glmRob     = c("broom::tidy", "broom::glance", "broom::augment"),

  # Regularized/glmnet models
  glmnet     = c("broom::tidy", "broom::glance"),
  cv.glmnet  = c("broom::tidy", "broom::glance"),

  # Nonlinear models
  nls        = c("broom::tidy", "broom::glance", "broom::augment"),
  nlrq       = c("broom::tidy", "broom::glance", "broom::augment"),

  # Mixed effects / multilevel
  lme        = c("broom::tidy", "broom::glance"),
  merMod     = c("broom::tidy", "broom::glance"),
  lmerMod    = c("broom::tidy", "broom::glance"),
  glmerMod   = c("broom::tidy", "broom::glance"),
  nlmerMod   = c("broom::tidy", "broom::glance"),

  # GAM models
  gam        = c("broom::tidy", "broom::glance", "broom::augment"),
  Gam        = c("broom::tidy", "broom::glance"),

  # Survival models
  coxph      = c("broom::tidy", "broom::glance", "broom::augment"),
  survreg    = c("broom::tidy", "broom::glance", "broom::augment"),
  aareg      = c("broom::tidy", "broom::glance"),
  cch        = c("broom::tidy", "broom::glance"),
  crr        = c("broom::tidy", "broom::glance"),
  survfit    = c("broom::tidy", "broom::glance"),
  survdiff   = c("broom::tidy", "broom::glance"),
  survexp    = c("broom::tidy", "broom::glance"),
  muhaz      = c("broom::tidy", "broom::glance"),
  pyears     = c("broom::tidy", "broom::glance"),

  # Time series / ARIMA
  Arima      = c("broom::tidy", "broom::glance"),
  acf        = "broom::tidy",
  spec       = "broom::tidy",
  ts         = "broom::tidy",
  zoo        = "broom::tidy",

  # Bayesian models
  stanfit    = c("broom::tidy", "broom::glance"),
  bamlss     = c("broom::tidy", "broom::glance"),

  # Panel data models
  plm        = c("broom::tidy", "broom::glance", "broom::augment"),

  # Count / GLM extensions
  negbin     = c("broom::tidy", "broom::glance"),
  speedglm   = c("broom::tidy", "broom::glance", "broom::augment"),

  # GEE models
  geeglm     = c("broom::tidy", "broom::glance"),

  # Ordinal models
  polr       = c("broom::tidy", "broom::glance", "broom::augment"),
  clm        = c("broom::tidy", "broom::glance", "broom::augment"),
  clmm       = c("broom::tidy", "broom::glance"),
  svyolr     = c("broom::tidy", "broom::glance"),

  # Multinomial models
  multinom   = c("broom::tidy", "broom::glance"),
  mlogit     = c("broom::tidy", "broom::glance", "broom::augment"),

  # Beta regression
  betareg    = c("broom::tidy", "broom::glance", "broom::augment"),

  # Cure regression
  crr        = c("broom::tidy", "broom::glance"),

  # Cluster analysis
  kmeans     = c("broom::tidy", "broom::glance", "broom::augment"),
  Mclust     = c("broom::tidy", "broom::glance", "broom::augment"),
  pam        = c("broom::tidy", "broom::glance", "broom::augment"),

  # PCA / dimension reduction
  prcomp     = c("broom::tidy", "broom::augment"),
  factanal   = c("broom::tidy", "broom::glance", "broom::augment"),

  # DEA / stochastic frontier
  felm       = c("broom::tidy", "broom::glance", "broom::augment"),
  ridgelm    = c("broom::tidy", "broom::glance"),

  # Quantile regression
  rq         = c("broom::tidy", "broom::glance", "broom::augment"),
  rqs        = c("broom::tidy", "broom::glance", "broom::augment"),

  # GMM
  gmm        = c("broom::tidy", "broom::glance"),

  # Instrumental variables
  ivreg      = c("broom::tidy", "broom::glance", "broom::augment"),

  # Meta-analysis
  rma        = c("broom::tidy", "broom::glance", "broom::augment"),

  # Network models
  ergm       = c("broom::tidy", "broom::glance"),
  btergm     = "broom::tidy",
  sarlm      = c("broom::tidy", "broom::glance", "broom::augment"),
  Sarlm      = c("broom::tidy", "broom::glance", "broom::augment"),

  # Marginal effects
  mfx        = c("broom::tidy", "broom::glance", "broom::augment"),
  betamfx    = c("broom::tidy", "broom::glance", "broom::augment"),
  logitmfx   = c("broom::tidy", "broom::glance", "broom::augment"),
  probitmfx  = c("broom::tidy", "broom::glance", "broom::augment"),
  negbinmfx  = c("broom::tidy", "broom::glance", "broom::augment"),
  poissonmfx = c("broom::tidy", "broom::glance", "broom::augment"),
  margins    = c("broom::tidy", "broom::glance", "broom::augment"),

  # Fixest models
  fixest     = c("broom::tidy", "broom::glance", "broom::augment"),

  # Survey models
  svyglm     = c("broom::tidy", "broom::glance"),

  # Latent class
  poLCA      = c("broom::tidy", "broom::glance", "broom::augment"),

  # Joint models
  mjoint     = c("broom::tidy", "broom::glance", "broom::augment"),

  # DRM / DRC
  drc        = c("broom::tidy", "broom::glance", "broom::augment"),

  # GARCH
  garch      = c("broom::tidy", "broom::glance"),

  # ANOVA extensions
  anova      = c("broom::tidy", "broom::glance"),
  aovlist    = "broom::tidy",

  # Coefficient tests
  coeftest   = c("broom::tidy", "broom::glance"),

  # Tests and diagnostics
  htest      = c("broom::tidy", "broom::glance", "broom::augment"),
  durbinWatsonTest = c("broom::tidy", "broom::glance"),

  # Confidence intervals
  confint.glht = "broom::tidy",
  glht        = "broom::tidy",
  summary.glht = "broom::tidy",

  # Bootstrapping
  boot       = "broom::tidy",

  # Statistical tests
  pairwise.htest = "broom::tidy",
  leveneTest = "broom::tidy",
  TukeyHSD   = "broom::tidy",

  # Table/matrix summaries
  table      = "broom::tidy",
  ftable     = "broom::tidy",
  confusionMatrix = "broom::tidy",

  # Variance models
  varest     = c("broom::tidy", "broom::glance"),
  lavaan     = c("broom::tidy", "broom::glance"),
  ref.grid   = "broom::tidy",
  emmGrid    = "broom::tidy",
  summary_emm = "broom::tidy",
  lsmobj     = "broom::tidy",

  # Binomial design
  binDesign  = c("broom::tidy", "broom::glance"),
  binWidth   = "broom::tidy",
  power.htest = "broom::tidy"
)

#' @importFrom utils getSrcref str
#' @importFrom grDevices dev.off pdf
#' @importFrom yaml read_yaml
#' @importFrom stats lm
NULL

#' Reset sandbox state (internal)
#' @keywords internal
.reset_last_sandbox <- function() {
  .resultcheck_env$last_sandbox <- NULL
}

test_that("setup_sandbox creates a temporary directory", {
  # Create a simple test file
  test_file <- tempfile(fileext = ".txt")
  writeLines("test content", test_file)
  on.exit(unlink(test_file))

  withr::with_dir(dirname(test_file), {
    # Setup sandbox
    sandbox <- setup_sandbox(basename(test_file))

    # Check that sandbox was created
    expect_s3_class(sandbox, "resultcheck_sandbox")
    expect_true(dir.exists(sandbox$path))
    expect_true(nchar(sandbox$id) > 0)

    # Clean up
    cleanup_sandbox(sandbox)
  })
})

test_that("setup_sandbox accepts no files argument to create an empty sandbox", {
  sandbox <- setup_sandbox()
  on.exit(cleanup_sandbox(sandbox), add = TRUE)

  expect_s3_class(sandbox, "resultcheck_sandbox")
  expect_true(dir.exists(sandbox$path))
})


test_that("setup_sandbox copies files with directory structure", {
  # Create test directory structure
  temp_root <- tempfile()
  dir.create(temp_root)
  dir.create(file.path(temp_root, "data"), recursive = TRUE)

  test_file1 <- file.path(temp_root, "data", "test1.txt")
  writeLines("content 1", test_file1)

  on.exit(unlink(temp_root, recursive = TRUE))

  # Setup sandbox with working directory set to temp_root
  withr::with_dir(temp_root, {
    sandbox <- setup_sandbox(c("data/test1.txt"))
    on.exit(cleanup_sandbox(sandbox), add = TRUE)

    # Check that file was copied with directory structure
    expect_true(file.exists(file.path(sandbox$path, "data", "test1.txt")))

    # Check content
    copied_content <- readLines(file.path(sandbox$path, "data", "test1.txt"))
    expect_equal(copied_content, "content 1")
  })
})


test_that("setup_sandbox handles custom temp base", {
  # Create custom temp base
  custom_base <- tempfile()
  dir.create(custom_base)
  on.exit(unlink(custom_base, recursive = TRUE))

  # Create a test file
  test_file <- tempfile(fileext = ".txt")
  writeLines("test", test_file)
  on.exit(unlink(test_file), add = TRUE)
  withr::with_dir(dirname(test_file), {
    # Setup sandbox with custom base
    sandbox <- setup_sandbox(basename(test_file), temp_base = custom_base)
    on.exit(cleanup_sandbox(sandbox), add = TRUE)

    # Check that sandbox is in custom base
    expect_true(grepl(custom_base, sandbox$path, fixed = TRUE))
  })
})


test_that("setup_sandbox warns about missing files", {
  expect_warning(
    sandbox <- setup_sandbox("nonexistent_file.txt"),
    "File not found"
  )

  # Clean up if sandbox was created
  if (exists("sandbox")) {
    cleanup_sandbox(sandbox)
  }
})


test_that("setup_sandbox stores last sandbox in package environment", {
  test_file <- tempfile(fileext = ".txt")
  writeLines("test", test_file)
  on.exit(unlink(test_file))
  withr::with_dir(dirname(test_file), {
    sandbox <- setup_sandbox(basename(test_file))
    on.exit(cleanup_sandbox(sandbox), add = TRUE)

    # Check that it's stored
    expect_identical(resultcheck:::.resultcheck_env$last_sandbox, sandbox)
  })
})


test_that("run_in_sandbox executes script in sandbox directory", {
  skip_if_not_installed("withr")

  # Create a test script
  script_file <- tempfile(fileext = ".R")
  writeLines(
    c(
      'writeLines("test output", "output.txt")',
      'saveRDS(getwd(), "wd.rds")'
    ),
    script_file
  )
  on.exit(unlink(script_file))

  # Create sandbox
  sandbox <- setup_sandbox(character(0))
  on.exit(cleanup_sandbox(sandbox), add = TRUE)

  # Run script in sandbox
  expect_true(run_in_sandbox(script_file, sandbox))

  # Check that output was created in sandbox
  expect_true(file.exists(file.path(sandbox$path, "output.txt")))

  # Check that script ran in sandbox directory
  saved_wd <- readRDS(file.path(sandbox$path, "wd.rds"))
  expect_equal(normalizePath(saved_wd), normalizePath(sandbox$path))
})


test_that("run_in_sandbox uses last sandbox by default", {
  skip_if_not_installed("withr")

  # Create a test script
  script_file <- tempfile(fileext = ".R")
  writeLines('writeLines("test", "output.txt")', script_file)
  on.exit(unlink(script_file))

  # Create sandbox (which becomes the last sandbox)
  sandbox <- setup_sandbox(character(0))
  on.exit(cleanup_sandbox(sandbox), add = TRUE)

  # Run without specifying sandbox
  expect_true(run_in_sandbox(script_file))

  # Check that output was created
  expect_true(file.exists(file.path(sandbox$path, "output.txt")))
})


test_that("run_in_sandbox suppresses messages and warnings", {
  skip_if_not_installed("withr")

  # Create a test script with messages and warnings
  script_file <- tempfile(fileext = ".R")
  writeLines(
    c(
      'message("This is a message")',
      'warning("This is a warning")'
    ),
    script_file
  )
  on.exit(unlink(script_file))

  # Create sandbox
  sandbox <- setup_sandbox(character(0))
  on.exit(cleanup_sandbox(sandbox), add = TRUE)

  # Run with suppression (default) - should not show messages/warnings
  result <- expect_silent(run_in_sandbox(script_file, sandbox))
  expect_true(result)
})


test_that("run_in_sandbox errors on missing script", {
  sandbox <- setup_sandbox(character(0))
  on.exit(cleanup_sandbox(sandbox), add = TRUE)

  expect_error(
    run_in_sandbox("nonexistent_script.R", sandbox),
    "Script file not found"
  )
})


test_that("run_in_sandbox errors without sandbox", {
  # Clear last sandbox
  resultcheck:::.reset_last_sandbox()

  script_file <- tempfile(fileext = ".R")
  writeLines('1 + 1', script_file)
  on.exit(unlink(script_file))

  expect_error(
    run_in_sandbox(script_file),
    "No sandbox specified"
  )
})


test_that("cleanup_sandbox removes directory", {
  test_file <- tempfile(fileext = ".txt")
  writeLines("test", test_file)
  on.exit(unlink(test_file))

  withr::with_dir(dirname(test_file), {
    sandbox <- setup_sandbox(basename(test_file))

    # Check directory exists
    expect_true(dir.exists(sandbox$path))

    # Clean up
    result <- cleanup_sandbox(sandbox)

    # Check directory is gone
    expect_false(dir.exists(sandbox$path))
    expect_true(result)
  })
})


test_that("cleanup_sandbox clears last_sandbox", {
  test_file <- tempfile(fileext = ".txt")
  writeLines("test", test_file)
  on.exit(unlink(test_file))

  withr::with_dir(dirname(test_file), {
    sandbox <- setup_sandbox(basename(test_file))

    # Verify it's stored
    expect_false(is.null(resultcheck:::.resultcheck_env$last_sandbox))

    # Clean up
    cleanup_sandbox(sandbox)

    # Check it's cleared
    expect_null(resultcheck:::.resultcheck_env$last_sandbox)
  })
})


test_that("cleanup_sandbox uses last sandbox by default", {
  test_file <- tempfile(fileext = ".txt")
  writeLines("test", test_file)
  on.exit(unlink(test_file))

  withr::with_dir(dirname(test_file), {
    sandbox <- setup_sandbox(basename(test_file))
    sandbox_path <- sandbox$path

    # Clean up without specifying sandbox
    cleanup_sandbox()

    # Check directory is gone
    expect_false(dir.exists(sandbox_path))
  })
})


test_that("cleanup_sandbox handles nonexistent directory gracefully", {
  test_file <- tempfile(fileext = ".txt")
  writeLines("test", test_file)
  on.exit(unlink(test_file))

  withr::with_dir(dirname(test_file), {
    sandbox <- setup_sandbox(basename(test_file))

    # Manually delete directory
    unlink(sandbox$path, recursive = TRUE)

    # Should warn but not error
    expect_warning(
      result <- cleanup_sandbox(sandbox),
      "does not exist"
    )
    expect_false(result)
  })
})


test_that("Full workflow: setup, run, cleanup", {
  skip_if_not_installed("withr")

  # Create test data file
  data_file <- tempfile(fileext = ".rds")
  saveRDS(data.frame(x = 1:5, y = 6:10), data_file)
  on.exit(unlink(data_file))

  withr::with_dir(dirname(data_file), {
    # Create test script
    script_file <- tempfile(fileext = ".R")
    writeLines(
      c(
        sprintf('data <- readRDS("%s")', basename(data_file)),
        'data$z <- data$x + data$y',
        'saveRDS(data, "output.rds")'
      ),
      script_file
    )
    on.exit(unlink(script_file), add = TRUE)

    # Full workflow
    sandbox <- setup_sandbox(basename(data_file))
    expect_true(run_in_sandbox(script_file, sandbox))

    # Verify output
    output_file <- file.path(sandbox$path, "output.rds")
    expect_true(file.exists(output_file))

    result <- readRDS(output_file)
    expect_equal(result$z, c(7, 9, 11, 13, 15))

    # Cleanup
    cleanup_sandbox(sandbox)
    expect_false(dir.exists(sandbox$path))
  })
})


test_that("setup_sandbox rejects absolute paths", {
  # Test Unix-style absolute path
  expect_error(
    setup_sandbox("/etc/passwd"),
    "Absolute paths are not allowed"
  )
  
  # Test another Unix-style absolute path
  expect_error(
    setup_sandbox("/tmp/test.txt"),
    "Absolute paths are not allowed"
  )
  
  # On Windows, test drive letter paths (will only trigger on Windows)
  if (.Platform$OS.type == "windows") {
    expect_error(
      setup_sandbox("C:/test.txt"),
      "Absolute paths are not allowed"
    )
    
    expect_error(
      setup_sandbox("D:\\test.txt"),
      "Absolute paths are not allowed"
    )
  }
})


test_that("setup_sandbox rejects path traversal attempts", {
  # Test basic path traversal
  expect_error(
    setup_sandbox("../etc/passwd"),
    "Path traversal"
  )
  
  # Test nested path traversal
  expect_error(
    setup_sandbox("data/../../etc/passwd"),
    "Path traversal"
  )
  
  # Test path traversal in middle
  expect_error(
    setup_sandbox("a/../b/file.txt"),
    "Path traversal"
  )
  
  # Test Windows-style path traversal
  expect_error(
    setup_sandbox("data\\..\\file.txt"),
    "Path traversal"
  )
})


test_that("setup_sandbox accepts legitimate filenames with double dots", {
  # Create test file with .. in filename
  temp_root <- tempfile()
  dir.create(temp_root)
  test_file <- file.path(temp_root, "file..txt")
  writeLines("content", test_file)
  on.exit(unlink(temp_root, recursive = TRUE))
  
  withr::with_dir(temp_root, {
    # Should accept filename with .. that is not a path component
    sandbox <- setup_sandbox("file..txt")
    expect_s3_class(sandbox, "resultcheck_sandbox")
    expect_true(file.exists(file.path(sandbox$path, "file..txt")))
    cleanup_sandbox(sandbox)
  })
  
  # Test nested path with filename containing ..
  temp_root2 <- tempfile()
  dir.create(temp_root2)
  dir.create(file.path(temp_root2, "subdir"))
  test_file2 <- file.path(temp_root2, "subdir", "data..csv")
  writeLines("csv,data", test_file2)
  on.exit(unlink(temp_root2, recursive = TRUE), add = TRUE)
  
  withr::with_dir(temp_root2, {
    # Should accept nested path with .. in filename
    sandbox <- setup_sandbox("subdir/data..csv")
    expect_s3_class(sandbox, "resultcheck_sandbox")
    expect_true(file.exists(file.path(sandbox$path, "subdir", "data..csv")))
    cleanup_sandbox(sandbox)
  })
})


test_that("setup_sandbox accepts valid relative paths", {
  # Create test files
  test_file <- tempfile(fileext = ".txt")
  writeLines("test content", test_file)
  on.exit(unlink(test_file))
  
  withr::with_dir(dirname(test_file), {
    # Simple basename should work
    sandbox <- setup_sandbox(basename(test_file))
    expect_s3_class(sandbox, "resultcheck_sandbox")
    cleanup_sandbox(sandbox)
  })
  
  # Test with subdirectory
  temp_root <- tempfile()
  dir.create(temp_root)
  dir.create(file.path(temp_root, "subdir"), recursive = TRUE)
  test_file2 <- file.path(temp_root, "subdir", "test.txt")
  writeLines("content", test_file2)
  on.exit(unlink(temp_root, recursive = TRUE), add = TRUE)
  
  withr::with_dir(temp_root, {
    # Relative path with subdirectory should work
    sandbox <- setup_sandbox("subdir/test.txt")
    expect_s3_class(sandbox, "resultcheck_sandbox")
    expect_true(file.exists(file.path(sandbox$path, "subdir", "test.txt")))
    cleanup_sandbox(sandbox)
  })
})


test_that("setup_sandbox copies entire directory recursively", {
  temp_root <- tempfile()
  dir.create(temp_root)
  dir.create(file.path(temp_root, "data"), recursive = TRUE)
  dir.create(file.path(temp_root, "data", "raw"), recursive = TRUE)
  writeLines("a,b\n1,2", file.path(temp_root, "data", "input.csv"))
  writeLines("x,y\n3,4", file.path(temp_root, "data", "raw", "other.csv"))
  on.exit(unlink(temp_root, recursive = TRUE))

  withr::with_dir(temp_root, {
    # Pass the directory name rather than individual file paths
    sandbox <- setup_sandbox("data")
    on.exit(cleanup_sandbox(sandbox), add = TRUE)

    expect_s3_class(sandbox, "resultcheck_sandbox")
    expect_true(file.exists(file.path(sandbox$path, "data", "input.csv")))
    expect_true(file.exists(file.path(sandbox$path, "data", "raw", "other.csv")))

    # Verify file contents were preserved
    copied <- readLines(file.path(sandbox$path, "data", "input.csv"))
    expect_equal(copied, c("a,b", "1,2"))
  })
})

test_that("example project helper creates consistent runnable files", {
  skip_if_not_installed("withr")

  with_example({
    expect_true(file.exists("_resultcheck.yml"))
    expect_true(file.exists("analysis.R"))
    expect_true(file.exists(file.path("tests", "_resultcheck_snaps", "analysis", "model.md")))
    expect_true(file.exists(file.path("tests", "_resultcheck_snaps", "analysis", "model_mismatch.md")))
    expect_true(file.exists(file.path("tests", "testthat", "test-analysis.R")))

    sandbox <- setup_sandbox()
    on.exit(cleanup_sandbox(sandbox), add = TRUE)
    expect_true(run_in_sandbox("analysis.R", sandbox))
  })
})


test_that("example project helper supports mismatch demos", {
  skip_if_not_installed("withr")

  with_example({
    sandbox <- setup_sandbox()
    on.exit(cleanup_sandbox(sandbox), add = TRUE)

    expect_error(
      run_in_sandbox("analysis.R", sandbox),
      "Snapshot differences found"
    )
  }, mismatch = TRUE)
})

---
title: "Automated Testing with GitHub Actions"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Automated Testing with GitHub Actions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette shows how to combine `resultcheck`, `testthat`, `renv`, and a
GitHub Actions workflow into a fully automated reproducibility pipeline.  The
goal is to make every push to your repository trigger a test run that fails
loudly whenever any analysis result drifts from its committed snapshot —
across all three major operating systems.

For a real-world example of this pattern in action, see the
[IMF replication repository](https://github.com/IMFPaper/IMF).

---

## Why each piece matters

| Tool | Role |
|------|------|
| `resultcheck` | Captures named snapshots of R objects; errors in CI when a snapshot changes |
| `testthat` | Test harness that runs the snapshots and reports failures |
| `renv` | Locks every package to an exact version so the environment is reproducible |
| GitHub Actions | Runs the test suite automatically on push/PR across Windows, macOS, and Linux |

Without `renv`, a routine package update could silently change a coefficient or
table.  Without snapshots, tests would not catch numerical drift.  Without
multi-OS CI, platform-specific floating-point differences would go unnoticed.

---

## Step 1 — Project layout

A minimal project that uses this workflow looks like:

```
myproject/
├── .Rprofile                   # auto-activates renv
├── renv.lock                   # locked package versions (committed)
├── renv/                       # renv internals (mostly gitignored)
├── _resultcheck.yml             # marks the project root
├── data/
│   └── panel_data.rds
├── code/
│   └── analysis.R              # your analysis script
├── tests/_resultcheck_snaps/     # committed snapshot files
│   └── analysis/
│       └── main_model.md
└── tests/
    └── testthat/
        └── test-analysis.R
```

The `_resultcheck.yml` file at the root can be empty — its presence is enough
for `find_root()` to locate the project:

```yaml
# _resultcheck.yml
```

---

## Step 2 — Initialise renv

Inside R, with your project open:

```r
install.packages("renv")
renv::init()
```

Install the packages your project needs, then snapshot the environment:

```r
renv::install(c("resultcheck", "testthat"))
# ... install any other packages your analysis uses ...
renv::snapshot()
```

Commit both `.Rprofile` and `renv.lock`.  The `renv/` folder should be
partially ignored according to renv's own `.gitignore` (created automatically
by `renv::init()`).

---

## Step 3 — Add snapshots to your analysis script

Call `resultcheck::snapshot()` on every object whose value matters for
reproducibility.  The first time you run the script interactively the snapshot
is saved; on all subsequent runs it is compared against the saved version.

```r
# code/analysis.R
data  <- readRDS("data/panel_data.rds")
model <- lm(y ~ x1 + x2, data = data)

resultcheck::snapshot(model,  "main_model")
resultcheck::snapshot(data,   "panel_data")

# ... continue writing outputs ...
```

Run the script interactively once to generate the `.md` snapshot files, review
them, then commit them to version control.

---

## Step 4 — Write a testthat test

```r
# tests/testthat/test-analysis.R
library(testthat)
library(resultcheck)

test_that("analysis produces stable results", {
  sandbox <- setup_sandbox("data")
  on.exit(cleanup_sandbox(sandbox), add = TRUE)

  # snapshot() inside run_in_sandbox() errors on any mismatch
  expect_true(run_in_sandbox("code/analysis.R", sandbox))
})
```

Run locally to confirm everything passes before pushing:

```r
testthat::test_dir("tests/testthat")
```

For package examples and quick demos, you can avoid writing into your current
project by wrapping code in `resultcheck::with_example({...})`, which creates a
temporary project in `tempdir()` and cleans it up automatically.

---

## Step 5 — GitHub Actions workflow

Create `.github/workflows/run-tests.yml`.  The key ingredients are:

* **Matrix strategy** — runs on Windows, macOS, and Ubuntu so
  platform-specific numerical differences are caught early.
* **OS-specific system libraries** — packages such as `ragg`, `xml2`, or
  `curl` need native libraries on Linux and macOS that are not required on
  Windows.
* **renv cache** — `r-lib/actions/setup-renv@v2` restores the `renv` cache
  between runs, avoiding re-installing hundreds of packages on every push.

```yaml
name: R Tests

on:
  push:
    branches: [ main ]
  pull_request:
    branches: [ main ]
  workflow_dispatch:

jobs:
  test:
    runs-on: ${{ matrix.config.os }}
    name: ${{ matrix.config.os }}

    strategy:
      fail-fast: false
      matrix:
        config:
          - {os: windows-latest}
          - {os: ubuntu-latest}
          - {os: macos-latest}

    env:
      GITHUB_PAT: ${{ secrets.GITHUB_TOKEN }}
      R_KEEP_PKG_SOURCE: yes

    steps:
      - name: Checkout repository
        uses: actions/checkout@v4

      - name: Setup R
        uses: r-lib/actions/setup-r@v2
        with:
          use-public-rspm: true

      # ── Linux system libraries ──────────────────────────────────────────
      - name: Install system dependencies (Linux)
        if: runner.os == 'Linux'
        run: |
          sudo apt-get update
          sudo apt-get install -y \
            libcurl4-openssl-dev \
            libssl-dev \
            libxml2-dev \
            libfontconfig1-dev \
            libharfbuzz-dev \
            libfribidi-dev \
            libfreetype6-dev \
            libpng-dev \
            libtiff5-dev \
            libjpeg-dev

      # ── macOS system libraries ───────────────────────────────────────────
      - name: Install system dependencies (macOS)
        if: runner.os == 'macOS'
        run: |
          set -euxo pipefail
          brew update
          brew install pkg-config libpng cairo freetype harfbuzz fribidi

          BREW_PREFIX="$(brew --prefix)"
          echo "SDKROOT=$(xcrun --sdk macosx --show-sdk-path)" >> $GITHUB_ENV
          echo "PATH=${BREW_PREFIX}/bin:${PATH}"               >> $GITHUB_ENV
          echo "PKG_CONFIG_PATH=${BREW_PREFIX}/lib/pkgconfig:$(brew --prefix libpng)/lib/pkgconfig" >> $GITHUB_ENV

          mkdir -p ~/.R
          PNG_CFLAGS="$(pkg-config --cflags libpng)"
          PNG_LIBS="$(pkg-config --libs   libpng)"
          {
            echo "CPPFLAGS += -I${BREW_PREFIX}/include"
            echo "LDFLAGS  += -L${BREW_PREFIX}/lib -Wl,-rpath,${BREW_PREFIX}/lib"
            echo "PKG_CPPFLAGS += ${PNG_CFLAGS}"
            echo "PKG_LIBS     += ${PNG_LIBS}"
          } >> ~/.R/Makevars

      # ── Restore renv cache (fast!) ───────────────────────────────────────
      - name: Restore renv packages
        uses: r-lib/actions/setup-renv@v2
        with:
          cache-version: 2

      # ── Run tests ────────────────────────────────────────────────────────
      - name: Run tests
        run: Rscript -e "testthat::test_dir('tests/testthat')"

      - name: Upload test artefacts on failure
        if: failure()
        uses: actions/upload-artifact@v4
        with:
          name: test-results-${{ matrix.config.os }}
          path: tests/testthat/*.Rout*
```

### How `setup-renv` caching works

`r-lib/actions/setup-renv@v2` reads your `renv.lock` file, computes a cache
key from its hash, and restores a previously saved cache of installed packages
before calling `renv::restore()`.  When the lock file has not changed, all
packages are served from the cache and `renv::restore()` completes in seconds
rather than minutes.

Increment `cache-version` (e.g. from `2` to `3`) whenever you want to force a
full re-install — for example after an OS upgrade or when debugging a strange
linking error.

---

## Step 6 — Snapshot lifecycle on CI

```
Developer                         CI runner
─────────                         ─────────
1. Edit analysis.R
2. Run interactively → snapshots
   generated / updated
3. Review diffs, accept changes
4. git add tests/_resultcheck_snaps/
   git commit && git push
                                  5. Workflow triggered
                                  6. renv::restore() (from cache)
                                  7. testthat::test_dir()
                                     └─ run_in_sandbox("code/analysis.R")
                                        └─ snapshot() in *testing mode*
                                           ✓ matches committed file → pass
                                           ✗ differs             → FAIL
```

CI never updates snapshots; it only enforces them.  To accept a legitimate
result change, always re-run the script interactively, confirm the diff, and
commit the updated `.md` files.

---

## Handling platform differences

When the same computation yields slightly different floating-point values on
different operating systems, use the mechanisms described in
`vignette("snapshot-tolerance")`:

* **`[ignored]` markers** — replace a volatile line in the snapshot file with
  the literal text `[ignored]`.  That line position is skipped on every
  platform.
* **`snapshot.precision`** — add a `precision` key to `_resultcheck.yml` to
  round all floating-point numbers before comparison:

```yaml
# _resultcheck.yml
snapshot:
  precision: 10
```

Either option lets CI pass on all platforms without losing the safety net on
the lines that do matter.

---

## Tips

* **Commit `renv.lock` but not `renv/library/`.**  The library is rebuilt from
  the lock file on each runner; only the lock file needs to be in version
  control.
* **Keep snapshots human-readable.**  The `.md` files produced by `snapshot()`
  are plain text and diff well in pull requests — reviewers can see at a glance
  whether a coefficient changed.
* **Pin your R version** in the workflow matrix (e.g. `r-version: '4.4.2'`) if
  minor R releases have ever changed your numerical results.
* **`fail-fast: false`** lets all three OS jobs run to completion even when one
  fails, giving you a full picture of where the discrepancy occurs.
* **`workflow_dispatch`** allows you to trigger the workflow manually from the
  GitHub Actions UI — useful for debugging without having to push a commit.

---
title: "FAQ"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{FAQ}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## What is the scope of the package?

`resultcheck` helps you detect when the **results of your analysis change unexpectedly**.

It does this by snapshotting key outputs (e.g. data, models, tables) and comparing them when your code is rerun.

For a full example workflow, see the [Get Started guide](resultcheck.html).

In short, `resultcheck` works by snapshotting key results and checking whether they change when your analysis is rerun.


### Levels of usage

`resultcheck` operates at two main levels:

1. **Object level**

```r
snapshot(model, "main_model")
```

This is typically used during **interactive analysis**.

When you are satisfied with a result, you “lock it in” with a snapshot.
On subsequent runs, you will be notified if it changes.

2. **Script/project level**

```r
run_in_sandbox("analysis.R", sandbox)
```
This is used for **automated checks**, for example in tests or CI.

It reruns your analysis in a clean environment and verifies that previously snapshotted results remain unchanged.

This helps detect unintended changes due to:

- code modifications (e.g. refactoring)
- updates from collaborators
- differences in environments (packages, R version, OS)

### Typical workflow

In practice, you might:

- run snapshots interactively while developing your analysis
- run tests (e.g. via [`testthat`](https://testthat.r-lib.org)) before committing changes
- optionally use CI (e.g. [GitHub Actions](renv-github-actions.html)
) to run checks automatically

## Can I use resultcheck for exploratory work?

Yes — but with some care.

Academic research and data analysis are often iterative and exploratory, and new analyses frequently build on earlier results. `resultcheck` can be useful during this process by helping you ensure that previously established results remain stable as you continue working.

You can add `snapshot()` calls gradually as your analysis evolves, and update them when changes are expected. This allows you to introduce checks incrementally rather than all at once.

In practice, `resultcheck` works best when used to `lock in results that you consider stable`, rather than for rapidly changing intermediate outputs. As your project matures, this makes it easier to ensure reproducibility and detect unintended changes.

Combined with version control, this also enables you to *refactor and reorganize your code with confidence*, knowing that you will be alerted if any key results change unexpectedly.

### Caveat: script and object names

`resultcheck` identifies snapshots based on the **script name** and **snapshot name**.

This means that:

- renaming a script, or
- changing the name passed to `snapshot()`

will cause `resultcheck` to treat it as a new snapshot, rather than a modification of an existing one.

In general, once you have decided to snapshot a result, it is best to keep:

- script names, and
- snapshot names

stable.

## Is resultcheck a version control system?

No — `resultcheck` does not track code history or file changes.

Instead, it checks whether the **results produced by your analysis** change when code is rerun.

A helpful way to think about it is:

> Git tracks *what* changed in your code
> resultcheck tracks *whether your results* changed

## How can collaborators track changes in snapshots over time?

`resultcheck` focuses on checking whether results have changed, not on tracking their history.

Information such as:

- when a snapshot was created  
- who created it  
- which commit it corresponds to  

is handled by version control systems such as Git.

In practice, we recommend committing snapshots *alongside* the code changes that produce them, and writing clear **commit messages** that explain *why* results changed (e.g. due to code refactoring, data updates, or dependency changes). This makes it much easier for collaborators to interpret differences later.

You can inspect snapshot history using Git tools such as:

```bash
git log tests/_resultcheck_snaps/
git blame tests/_resultcheck_snaps/analysis/model.md
```

These commands show who changed what and when. Many code editors (e.g. VS Code, Positron) provide a *graphical* interface for this, allowing you to **trace each line of a snapshot** back to the commit that introduced it.

This makes it possible to understand changes at a very fine-grained level, while keeping snapshot files clean and focused on results. In this workflow, `resultcheck` detects changes in results, and Git explains them.

## Do I need to save models as `.rds` or `.rda` files to use resultcheck?

No.

You do **not** need to manually save anything to disk.

`resultcheck` snapshots objects directly as `.md` files, for example:

```r
model <- lm(mpg ~ wt, data = mtcars)
snapshot(model, "main_model")
```

This will create a snapshot file at: `tests/_resultcheck_snaps/[your_script_name]/main_model.md`.

## How does resultcheck differ from manually comparing the `.rds` or `.rda` files?

`.rds` and `.rda` files are binary formats, which can be large and difficult to compare directly.

In contrast, resultcheck stores snapshots as `.md` files, which are:

- human-readable
- easy to compare (e.g. with Git diffs)
- version control-friendly

We recommend committing these snapshot files to **version control**. This allows you to:

- see how results change across revisions
- collaborate more effectively
- automatically detect unintended changes via tests or CI

## Does resultcheck track changes over time?

No — it compares against the **current snapshot**.

It does not maintain a full history like version control.

If you want history, use Git — `resultcheck` complements it.

## What about large projects — is the sandbox expensive?

`setup_sandbox()` copies files into a temporary directory.

For very large projects, this can be expensive in time and disk space.

Current recommendations:

- include only necessary files
- avoid copying large unused directories

Future versions may include:

- size warnings
- more selective copying

## Is find_root() the same as here::here()?

Not exactly.

`find_root()` is an internal helper used by resultcheck to locate the project root. It is primarily used to identify which files (e.g. data, saved objects) should be copied into the sandbox when running analyses in a clean environment.

Unlike `here::here()`, `find_root()` will **error if no project root is found**, rather than guessing a path. This is intentional, to avoid silently writing or reading files from the wrong location.

In most cases, you should **not** need to call `find_root()` directly.

For constructing file paths in your own analysis code, we recommend using `here::here()` (or equivalent tools), as it is more flexible and designed for everyday use.

## Can I see exactly what changed in a model?

Currently, `resultcheck` reports differences as text diffs.

For example:

```r
Warning: 
Snapshot differences found for: model
File: E:/OneDrive/Desktop/ark-test/tests/_resultcheck_snaps/resultcheck-test/model.md

Differences:
old[2:8] vs new[2:8]
  ""
  "## List Structure"
  "List of 12"
- " $ coefficients : Named num [1:2] 40.03 -5.34"
+ " $ coefficients : Named num [1:2] 37.29 -5.34"
  "  ..- attr(*, \"names\")= chr [1:2] \"(Intercept)\" \"wt\""
  " $ residuals    : Named num [1:32] -2.28 -0.92 -2.09 1.3 -0.2 ..."
  "  ..- attr(*, \"names\")= chr [1:32] \"Mazda RX4\" \"Mazda RX4 Wag\" \"Datsun 710\" \"Hornet 4 Drive\" ..."

Update snapshot? (y/n): 
```

More structured comparisons are a possible future extension.

---
title: "Handling Platform Differences in Snapshots"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Handling Platform Differences in Snapshots}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(resultcheck)
```

## The problem: floating-point noise across platforms

Numerical computations in R can produce results that differ at very high decimal
places depending on the platform, compiler, or the version of libraries like
BLAS/LAPACK.  For example, a centering attribute from `scale()` might be
`2.41e-17` on one machine and `1.63e-17` on another — both are effectively zero,
but a naïve snapshot comparison treats them as different.

`resultcheck` provides two complementary mechanisms to handle this:

1. **`[ignored]` line markers** — mark specific lines in a snapshot file as
   known discrepancies that should always be skipped.
2. **`snapshot.precision` in `_resultcheck.yml`** — round all floating-point
   numbers to a fixed number of decimal places before comparing or storing
   snapshots.

---

## Feature 1: `[ignored]` line markers

### How it works

Open any `.md` snapshot file under `tests/_resultcheck_snaps/` and replace the
entire content of a volatile line with the literal text `[ignored]`.  On every
subsequent run, that line position is masked before the comparison takes place,
so differences there will never trigger a failure.  When you update a snapshot
interactively, the `[ignored]` markers are automatically preserved in the new
file.

### Step-by-step example

Suppose your `PCA.R` script produces a snapshot that includes lines like:

```
  ..- attr(*, "scaled:center")= num 2.41e-17
  ..- attr(*, "scaled:center")= num 2.98e-17
```

and these values fluctuate slightly between platforms.  Open the snapshot file
at `tests/_resultcheck_snaps/PCA/panel_data_pca.md` and change those specific
lines to:

```
[ignored]
[ignored]
```

Save the file.  The snapshot now looks like this (abbreviated):

```markdown
# Snapshot: data.frame

## Structure
'data.frame':  1071 obs. of  30 variables:
 ...
 $ us : num [1:1071, 1] NA -0.423 -0.697 ...
[ignored]
  ..- attr(*, "scaled:scale")= num 1.22
 $ eu : num [1:1071, 1] NA -0.783 -0.711 ...
[ignored]
  ..- attr(*, "scaled:scale")= num 1.37
```

From now on, `run_in_sandbox("code/PCA.R", sandbox)` will not raise an error
regardless of what those two lines contain in any new run, while every other
line continues to be verified.

### Rules

* The **entire line** must be `[ignored]` (leading/trailing whitespace is
  stripped before the check, so indentation does not matter).
* Lines are matched by **position** within the snapshot file, so adding or
  removing other lines shifts the positions.  If the overall structure of the
  output changes materially, update the snapshot interactively first, then
  re-apply the `[ignored]` markers to the new file.
* `[ignored]` markers are **preserved automatically** when you update a
  snapshot: `resultcheck` copies each `[ignored]` line from the stored snapshot
  into the freshly generated one before writing it to disk.

---

## Feature 2: Numeric precision in `_resultcheck.yml`

### How it works

Add a `snapshot.precision` key to the `_resultcheck.yml` file at the root of
your project.  Before storing or comparing a snapshot, every floating-point
number in the text representation is rounded to that many decimal places.
Integer-like tokens (e.g. index ranges such as `[1:1071]`) are left unchanged.

```yaml
# _resultcheck.yml
snapshot:
  precision: 10
```

With `precision: 10`, both `2.41e-17` and `1.63e-17` round to `0`, so the
snapshots compare as equal.

### Step-by-step example

1. Add `precision` to `_resultcheck.yml`:

   ```yaml
   snapshot:
     precision: 10
   ```

2. Delete the outdated snapshot (or run interactively and accept the update
   when prompted) so that the stored file is regenerated with rounded values.

3. Re-run the script via `run_in_sandbox()`.  The comparison will now succeed
   whenever the effective difference is smaller than `0.5 × 10^{-precision}`.

### Choosing a precision value

| Situation | Suggested precision |
|-----------|-------------------|
| Machine-epsilon noise (`1e-15` to `1e-17`) | 10 |
| Monte-Carlo / bootstrap variance | 4–6 |
| Monetary amounts (cents) | 2 |

A higher value is more sensitive (closer to exact comparison); a lower value is
more forgiving.

### Backward compatibility

If your project already has snapshot files stored before `precision` was
configured, `resultcheck` also rounds the **stored** snapshot numbers during
comparison.  This means setting `precision` will immediately suppress
floating-point noise in existing snapshots without needing to regenerate them.

---

## Using both features together

The two features are independent and compose naturally:

* Use **`[ignored]`** when a line's value is entirely unreliable across runs
  (e.g. a random seed, a memory address, or a platform-specific constant that
  carries no scientific meaning).
* Use **`precision`** when values are meaningful but may differ in the last few
  decimal places due to floating-point arithmetic.

For the PCA example in the problem statement, either approach resolves the
failure.  Using `precision: 10` is the lower-maintenance option because it
requires no manual edits to snapshot files; `[ignored]` offers surgical control
when only a handful of lines are volatile while the rest must be exact.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{find_root}
\alias{find_root}
\title{Find Project Root Directory}
\usage{
find_root(start_path = NULL)
}
\arguments{
\item{start_path}{Optional. The directory to start searching from.
If NULL (default), uses the current working directory or the stored
original working directory if in a sandbox.}
}
\value{
The path to the project root directory.
}
\description{
Finds the root directory of the current R project using various heuristics.
The function searches for markers like \code{_resultcheck.yml} (preferred),
\code{resultcheck.yml} (legacy), \code{.Rproj} files,
or a \code{.git} directory. When running inside a sandbox created by
\code{setup_sandbox()}, it will search from the original working directory.
}
\examples{
with_example({
  root <- find_root()
  print(root)
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sandbox.R
\name{cleanup_sandbox}
\alias{cleanup_sandbox}
\title{Clean Up a Sandbox Environment}
\usage{
cleanup_sandbox(sandbox = NULL, force = TRUE)
}
\arguments{
\item{sandbox}{Optional. A sandbox object created by \code{setup_sandbox()}.
If NULL (default), cleans up the most recently created sandbox.}

\item{force}{Logical. If TRUE (default), removes directory even if it
contains files.}
}
\value{
Logical indicating success (invisible).
}
\description{
Removes a sandbox directory and all its contents. This should be called
after testing is complete to free up disk space.
}
\examples{
with_example({
  sandbox <- setup_sandbox()
  cleanup_sandbox(sandbox)
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{get_start_path_for_find_root}
\alias{get_start_path_for_find_root}
\title{Get start path for find_root}
\usage{
get_start_path_for_find_root()
}
\value{
Character path to start searching from
}
\description{
Helper function to determine the starting path for find_root().
Checks for stored sandbox WD first, then falls back to getwd().
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.reset_last_sandbox}
\alias{.reset_last_sandbox}
\title{Reset sandbox state (internal)}
\usage{
.reset_last_sandbox()
}
\description{
Reset sandbox state (internal)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{snapshot}
\alias{snapshot}
\title{Interactive Snapshot Testing}
\usage{
snapshot(value, name, script_name = NULL, method = NULL)
}
\arguments{
\item{value}{The R object to snapshot (e.g., plot, table, model output).}

\item{name}{Character. A descriptive name for this snapshot.}

\item{script_name}{Optional. The name of the script creating the snapshot.
If NULL, auto-detects via \code{rstudioapi::getSourceEditorContext()$path}
(in RStudio), falling back to the call stack, then to \code{"interactive"}.}

\item{method}{Optional function or non-empty list of functions used to
serialize \code{value}. Functions are executed in order and each section
header is taken from the method expression or list name. If omitted,
method defaults are resolved in this order:
\code{snapshot.method_by_class} (matched by object class), then
\code{snapshot.method}, then
\code{list(print = base::print, str = utils::str)}.}
}
\value{
Invisible TRUE if snapshot matches or was updated.
In testing mode, throws an error if snapshot is missing or doesn't match.
}
\description{
Creates or updates a snapshot of an R object for interactive analysis.
On first use, saves the object to a human-readable snapshot file (.md).
On subsequent uses, compares the current object to the saved snapshot.
}
\details{
In interactive mode (default), prompts the user to update if differences
are found and emits a warning. In testing mode (inside testthat or
run_in_sandbox), throws an error if snapshot doesn't exist or doesn't match.

Snapshots are stored under \code{tests/_resultcheck_snaps/} by default,
organized by script name, and configurable via \code{snapshot.dir} in
\code{_resultcheck.yml}. Method defaults can be configured via
\code{snapshot.method} and class-specific defaults via
\code{snapshot.method_by_class}. Optional class defaults can also be loaded
from an R file using \code{snapshot.method_defaults_file}. Method strings in
config (for example \code{"print + str"} or \code{"stats::coef"}) are
resolved to callable functions. In config expressions, \code{"+"} is treated
as the method delimiter.

Built-in class defaults (loaded from inst/extdata/snapshot-method-defaults.R)
use broom functions for many statistical model classes. The default method
is typically broom::tidy, with broom::glance and/or broom::augment
added where supported (per the broom available-methods table at
https://broom.tidymodels.org/articles/available-methods.html).
}
\examples{
with_example({
  model <- stats::lm(mpg ~ wt, data = datasets::mtcars)
  snapshot(model, "model_default", script_name = "analysis")
  snapshot(model, "model_multi", script_name = "analysis",
           method = list(summary = summary, print = print))
  snapshot(model, "model_print", script_name = "analysis", method = print)
  snapshot(model, "model_ns", script_name = "analysis", method = stats::coef)
  snapshot(model, "model_length", script_name = "analysis", method = length)
})

with_example({
  sandbox <- setup_sandbox()
  on.exit(cleanup_sandbox(sandbox), add = TRUE)
  run_in_sandbox("analysis.R", sandbox)
})

if (interactive()) with_example({
  sandbox <- setup_sandbox()
  on.exit(cleanup_sandbox(sandbox), add = TRUE)
  run_in_sandbox("analysis.R", sandbox)
}, mismatch = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sandbox.R
\name{with_example}
\alias{with_example}
\title{Run Code Inside a Temporary Example Project}
\usage{
with_example(code, mismatch = FALSE)
}
\arguments{
\item{code}{Code to evaluate inside the temporary example project.}

\item{mismatch}{Logical. If TRUE, replaces the active snapshot with a
mismatched version before evaluating \code{code}.}
}
\value{
The value of \code{code}.
}
\description{
Creates a self-contained example project under \code{tempdir()}, including:
\itemize{
\item \code{_resultcheck.yml} (project root marker)
\item \code{analysis.R} with \code{snapshot(model, "model")}
\item matching and mismatched snapshot files
\item \code{tests/testthat/test-analysis.R}
}
then temporarily sets the working directory to that project while
evaluating \code{code}.
}
\examples{
with_example({
  root <- find_root()
  print(root)
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sandbox.R
\name{.get_project_root}
\alias{.get_project_root}
\title{Get project root for sandbox operations}
\usage{
.get_project_root()
}
\value{
Character path to project root or current working directory
}
\description{
Internal helper to get project root, falling back to current directory
if root cannot be determined.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{get_snapshot_path}
\alias{get_snapshot_path}
\title{Get Snapshot File Path}
\usage{
get_snapshot_path(name, script_name = NULL, ext = "md")
}
\arguments{
\item{name}{Character. The name of the snapshot (without extension).}

\item{script_name}{Optional. The name of the script file creating the snapshot.
If NULL, detects from the active RStudio source editor via
\code{rstudioapi::getSourceEditorContext()$path}, falling back to the call
stack, then to \code{"interactive"}.}

\item{ext}{Character. The file extension for the snapshot file (default: "md").}
}
\value{
The full path to the snapshot file.
}
\description{
Constructs the path to a snapshot file within the project's snapshot directory.
Snapshot files are stored under \code{tests/_resultcheck_snaps/} by default,
organized by script name. This location can be overridden with
\code{snapshot.dir} in \code{_resultcheck.yml}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{round_snapshot_numbers}
\alias{round_snapshot_numbers}
\title{Round Floating-Point Numbers in Snapshot Text}
\usage{
round_snapshot_numbers(text, digits)
}
\arguments{
\item{text}{Character vector of snapshot text lines.}

\item{digits}{Integer; number of decimal places to keep (passed to
\code{round()}).  Must be a finite integer.}
}
\value{
Character vector with floating-point numbers rounded.
}
\description{
Replaces every floating-point literal (decimal or scientific notation) found
in a character vector of snapshot lines with its value rounded to
\code{digits} decimal places.  Integer literals that contain no decimal
point and no exponent are left untouched, so index ranges such as
\code{[1:1071]} are never modified.
}
\details{
\code{digits} is passed directly to \code{\link[base]{round}}: 0 rounds to
the nearest integer, negative values round to the left of the decimal point.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{compare_snapshot_text}
\alias{compare_snapshot_text}
\title{Compare Two Snapshot Values}
\usage{
compare_snapshot_text(old_text, new_text, precision = NULL)
}
\arguments{
\item{old_text}{Character vector with old snapshot text.}

\item{new_text}{Character vector with new snapshot text.}

\item{precision}{Optional integer.  When non-\code{NULL}, both texts are
rounded to this many decimal places before comparison (see
\code{\link{round_snapshot_numbers}}).  Useful for ignoring
floating-point noise introduced by platform differences.}
}
\value{
A character vector of differences, or NULL if identical.
}
\description{
Compares two serialized snapshots and returns differences.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{serialize_value}
\alias{serialize_value}
\title{Serialize Value to Human-Readable Text}
\usage{
serialize_value(value, methods = NULL, use_class_defaults = TRUE)
}
\arguments{
\item{value}{The R object to serialize.}

\item{methods}{A function or a non-empty list of functions applied in order.
Defaults to \code{list(print = base::print, str = utils::str)}.}
}
\value{
A character vector with the text representation.
}
\description{
Converts an R object to a human-readable text representation for snapshots.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{read_resultcheck_config}
\alias{read_resultcheck_config}
\title{Read resultcheck Configuration}
\usage{
read_resultcheck_config()
}
\value{
A named list of configuration values, or an empty list.
}
\description{
Reads configuration settings from the \code{_resultcheck.yml} file located
at the project root (falling back to legacy \code{resultcheck.yml} if
needed). Returns an empty list if neither file exists or parsing fails.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{mask_ignored_lines}
\alias{mask_ignored_lines}
\title{Mask \code{[ignored]} Lines in New Snapshot Text}
\usage{
mask_ignored_lines(old_text, new_text)
}
\arguments{
\item{old_text}{Character vector of the stored snapshot lines.}

\item{new_text}{Character vector of the freshly generated snapshot lines.}
}
\value{
\code{new_text} with \code{[ignored]} substituted at matching
positions.
}
\description{
Any line in \code{old_text} that equals \code{"[ignored]"} (after trimming
whitespace) causes the corresponding line in \code{new_text} to be replaced
with \code{"[ignored]"}, so that known-volatile lines never trigger a
snapshot failure.  Lines beyond the shorter vector are left unchanged.
}
\details{
This helper is used both during comparison (so the lines are skipped) and
when writing an updated snapshot (so the markers are preserved).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{normalize_snapshot_text}
\alias{normalize_snapshot_text}
\title{Normalize Snapshot Text Before Comparison}
\usage{
normalize_snapshot_text(text)
}
\arguments{
\item{text}{Character vector with serialized snapshot lines.}
}
\value{
Character vector with normalized snapshot lines.
}
\description{
Replaces volatile environment representations (for example memory addresses
in \code{.Environment} attributes) with a stable placeholder so snapshots remain
comparable across different execution contexts.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{is_testing}
\alias{is_testing}
\title{Detect Testing Context}
\usage{
is_testing()
}
\value{
Logical indicating if in testing mode.
}
\description{
Determines if code is running in a testing context (sandbox via run_in_sandbox).
This is used to change snapshot() behavior: interactive mode prompts for updates,
while testing mode throws errors on mismatch.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sandbox.R
\name{run_in_sandbox}
\alias{run_in_sandbox}
\title{Run Code in a Sandbox Environment}
\usage{
run_in_sandbox(
  script_path,
  sandbox = NULL,
  suppress_messages = TRUE,
  suppress_warnings = TRUE,
  capture_output = TRUE
)
}
\arguments{
\item{script_path}{Path to the R script to execute.}

\item{sandbox}{Optional. A sandbox object created by \code{setup_sandbox()}.
If NULL (default), uses the most recently created sandbox.}

\item{suppress_messages}{Logical. Whether to suppress messages (default: TRUE).}

\item{suppress_warnings}{Logical. Whether to suppress warnings (default: TRUE).}

\item{capture_output}{Logical. Whether to capture output (default: TRUE).}
}
\value{
Invisible TRUE on successful execution.
}
\description{
Executes an R script within a sandbox directory, suppressing messages,
warnings, and graphical output. This is useful for testing empirical
analysis scripts without polluting the console or creating unwanted plots.
}
\examples{
with_example({
  sandbox <- setup_sandbox()
  on.exit(cleanup_sandbox(sandbox), add = TRUE)
  run_in_sandbox("analysis.R", sandbox)
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sandbox.R
\name{setup_sandbox}
\alias{setup_sandbox}
\title{Setup a Sandbox Environment for Testing}
\usage{
setup_sandbox(files = NULL, temp_base = NULL)
}
\arguments{
\item{files}{Character vector of relative file or directory paths to copy
to the sandbox. Leave as \code{NULL} (default) to create an empty sandbox.
Paths are resolved relative to the project root (found
using \code{find_root()}); if the project root cannot be determined the
current working directory is used.  When a path refers to a directory,
the entire directory is copied recursively.  Absolute paths and path
traversal attempts (e.g., \code{..}) are rejected for security.
Snapshot files do \emph{not} need to be listed here: \code{snapshot()}
always reads snapshots from the project root, not from the sandbox.}

\item{temp_base}{Optional. Custom location for the temporary directory.
If NULL (default), uses \code{tempfile()}.}
}
\value{
A list with class "resultcheck_sandbox" containing:
\item{path}{The path to the created temporary directory}
\item{id}{A unique timestamp-based identifier for this sandbox}
}
\description{
Creates a temporary directory and copies specified files and/or directories
into it while preserving their path structure. This is useful for testing
empirical analysis scripts in isolation.
}
\examples{
with_example({
  sandbox <- setup_sandbox()
  print(sandbox$path)
  cleanup_sandbox(sandbox)
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{detect_script_name}
\alias{detect_script_name}
\title{Detect the calling script name}
\usage{
detect_script_name()
}
\value{
Character string with the detected script basename (without path).
}
\description{
Tries to identify the script file that is calling snapshot().
Uses \code{rstudioapi::getSourceEditorContext()$path} when running in RStudio,
falling back to walking the call stack for source references, and finally
to \code{"interactive"}.
}
\keyword{internal}
